# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os
import socket
import ssl
import sys
import typing

import pytest

import spnego
import spnego._credssp
import spnego._gss
import spnego._sspi
import spnego.channel_bindings
import spnego.iov
import spnego.tls
from spnego._context import (
    IOVUnwrapResult,
    IOVWrapResult,
    UnwrapResult,
    WinRMWrapResult,
    WrapResult,
)
from spnego._credssp_structures import TSPasswordCreds
from spnego._ntlm_raw.crypto import lmowfv1, ntowfv1
from spnego._spnego import NegTokenResp, unpack_token
from spnego.exceptions import (
    InvalidCredentialError,
    InvalidTokenError,
    NoCredentialError,
    SpnegoError,
)


def _message_test(client: spnego.ContextProxy, server: spnego.ContextProxy) -> None:
    # Client wrap
    plaintext = os.urandom(32)

    c_wrap_result = client.wrap(plaintext)

    assert isinstance(c_wrap_result, WrapResult)
    assert c_wrap_result.encrypted
    assert c_wrap_result.data != plaintext

    # Server unwrap
    s_unwrap_result = server.unwrap(c_wrap_result.data)

    assert isinstance(s_unwrap_result, UnwrapResult)
    assert s_unwrap_result.data == plaintext
    assert s_unwrap_result.encrypted
    assert s_unwrap_result.qop == 0

    # Server wrap
    plaintext = os.urandom(17)

    s_wrap_result = server.wrap(plaintext)

    assert isinstance(s_wrap_result, WrapResult)
    assert s_wrap_result.encrypted
    assert s_wrap_result.data != plaintext

    # Client unwrap
    c_unwrap_result = client.unwrap(s_wrap_result.data)

    assert isinstance(c_unwrap_result, UnwrapResult)
    assert c_unwrap_result.data == plaintext
    assert c_unwrap_result.encrypted
    assert c_unwrap_result.qop == 0

    # CredSSP doesnt' have signature methods
    if isinstance(client, spnego._credssp.CredSSPProxy):
        return

    # Client sign, server verify
    plaintext = os.urandom(3)

    c_sig = client.sign(plaintext)
    server.verify(plaintext, c_sig)

    # Server sign, client verify
    plaintext = os.urandom(9)

    s_sig = server.sign(plaintext)
    client.verify(plaintext, s_sig)

    # Can only continue if we are not testing Kerberos, or we are testing Kerberos and GSSAPI is available.
    if client.negotiated_protocol == "kerberos" and not client.iov_available():
        return

    plaintext = os.urandom(16)
    c_winrm_wrap_result = client.wrap_winrm(plaintext)
    assert isinstance(c_winrm_wrap_result, WinRMWrapResult)
    assert isinstance(c_winrm_wrap_result.header, bytes)
    assert isinstance(c_winrm_wrap_result.data, bytes)
    assert isinstance(c_winrm_wrap_result.padding_length, int)

    s_winrm_unwrap_result = server.unwrap_winrm(c_winrm_wrap_result.header, c_winrm_wrap_result.data)
    assert s_winrm_unwrap_result == plaintext

    plaintext = os.urandom(16)
    s_winrm_wrap_result = server.wrap_winrm(plaintext)
    assert isinstance(s_winrm_wrap_result, WinRMWrapResult)
    assert isinstance(s_winrm_wrap_result.header, bytes)
    assert isinstance(s_winrm_wrap_result.data, bytes)
    assert isinstance(s_winrm_wrap_result.padding_length, int)

    c_winrm_unwrap_result = client.unwrap_winrm(s_winrm_wrap_result.header, s_winrm_wrap_result.data)
    assert c_winrm_unwrap_result == plaintext

    # Can only continue if using Kerberos auth and IOV is available
    if client.negotiated_protocol == "ntlm" or not client.iov_available():
        return

    plaintext = os.urandom(16)
    c_iov_wrap_res = client.wrap_iov([spnego.iov.BufferType.header, plaintext, spnego.iov.BufferType.padding])
    assert isinstance(c_iov_wrap_res, IOVWrapResult)
    assert c_iov_wrap_res.encrypted
    assert len(c_iov_wrap_res.buffers) == 3
    assert c_iov_wrap_res.buffers[1].data != plaintext

    s_iov_unwrap_res = server.unwrap_iov(c_iov_wrap_res.buffers)
    assert isinstance(s_iov_unwrap_res, IOVUnwrapResult)
    assert s_iov_unwrap_res.encrypted
    assert s_iov_unwrap_res.qop == 0
    assert len(s_iov_unwrap_res.buffers) == 3
    assert s_iov_unwrap_res.buffers[1].data == plaintext

    plaintext = os.urandom(16)
    s_iov_wrap_res = server.wrap_iov([spnego.iov.BufferType.header, plaintext, spnego.iov.BufferType.padding])
    assert isinstance(s_iov_wrap_res, IOVWrapResult)
    assert s_iov_wrap_res.encrypted
    assert len(s_iov_wrap_res.buffers) == 3
    assert s_iov_wrap_res.buffers[1].data != plaintext

    c_iov_unwrap_res = client.unwrap_iov(s_iov_wrap_res.buffers)
    assert isinstance(c_iov_unwrap_res, IOVUnwrapResult)
    assert c_iov_unwrap_res.encrypted
    assert c_iov_unwrap_res.qop == 0
    assert len(c_iov_unwrap_res.buffers) == 3
    assert c_iov_unwrap_res.buffers[1].data == plaintext


def _ntlm_test(
    client: spnego.ContextProxy,
    server: spnego.ContextProxy,
    test_session_key: bool = True,
    channel_bindings: typing.Optional[spnego.channel_bindings.GssChannelBindings] = None,
) -> None:
    assert not client.complete
    assert not server.complete

    # Build negotiate msg
    negotiate = client.step(channel_bindings=channel_bindings)

    assert isinstance(negotiate, bytes)
    assert not client.complete
    assert not server.complete

    # Process negotiate msg
    challenge = server.step(negotiate, channel_bindings=channel_bindings)

    assert isinstance(challenge, bytes)
    assert not client.complete
    assert not server.complete

    # Process challenge and build authenticate
    authenticate = client.step(challenge, channel_bindings=channel_bindings)

    assert isinstance(authenticate, bytes)
    if test_session_key:
        assert isinstance(client.session_key, bytes)

    assert client.complete
    assert not server.complete

    # Process authenticate
    auth_response = server.step(authenticate, channel_bindings=channel_bindings)

    assert auth_response is None
    if test_session_key:
        assert isinstance(client.session_key, bytes)
        assert isinstance(server.session_key, bytes)

    assert client.complete
    assert server.complete

    assert client.negotiated_protocol == "ntlm"
    assert server.negotiated_protocol == "ntlm"


def test_invalid_protocol():
    expected = "Invalid protocol specified 'fake', must be kerberos, negotiate, or ntlm"

    with pytest.raises(ValueError, match=expected):
        spnego.client(None, None, protocol="fake")

    with pytest.raises(ValueError, match=expected):
        spnego.server(protocol="fake")


def test_protocol_not_supported():
    with pytest.raises(ValueError, match="Protocol kerberos is not available"):
        spnego.client(None, None, protocol="kerberos", options=spnego.NegotiateOptions.use_ntlm)


def test_no_valid_credential_available_single_available_protocol():
    credentials: typing.List[spnego.Credential] = [
        spnego.KerberosCCache("FILE:test"),
        spnego.KerberosKeytab("principal", "keytab"),
    ]
    with pytest.raises(
        NoCredentialError, match="A credential for ntlm is needed but only found credentials for kerberos"
    ):
        spnego.client(credentials, protocol="ntlm", options=spnego.NegotiateOptions.use_ntlm)


def test_no_valid_credential_available_multiple_available_protocol():
    credentials: typing.List[spnego.Credential] = [
        spnego.NTLMHash("username", "lm_hash", "nt_hash"),
        spnego.KerberosKeytab("principal", "keytab"),
    ]
    with pytest.raises(
        NoCredentialError, match="A credential for credssp is needed but only found credentials for kerberos, ntlm"
    ):
        spnego.client(credentials, protocol="credssp")


# Negotiate scenarios


def test_negotiate_with_kerberos(kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    c = spnego.client(
        kerb_cred.user_princ, None, hostname=socket.getfqdn(), options=spnego.NegotiateOptions.use_negotiate
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_negotiate)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    token1 = c.step()
    assert isinstance(token1, bytes)

    token2 = s.step(token1)
    assert isinstance(token2, bytes)

    token3 = c.step(token2)
    assert token3 is None

    # Make sure it reports the right protocol
    assert c.negotiated_protocol == "kerberos"
    assert s.negotiated_protocol == "kerberos"

    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr & spnego.ContextReq.mutual_auth
    assert s.context_attr & spnego.ContextReq.mutual_auth

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)

    token2 = s.step(token1)
    assert isinstance(token2, bytes)

    token3 = c.step(token2)
    assert token3 is None

    # Make sure it reports the right protocol
    assert c.negotiated_protocol == "kerberos"
    assert s.negotiated_protocol == "kerberos"

    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr & spnego.ContextReq.mutual_auth
    assert s.context_attr & spnego.ContextReq.mutual_auth

    _message_test(c, s)


def test_negotiate_with_kerberos_no_integrity(kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    c = spnego.client(
        kerb_cred.user_princ,
        None,
        hostname=socket.getfqdn(),
        options=spnego.NegotiateOptions.use_negotiate,
        context_req=spnego.ContextReq.mutual_auth | spnego.ContextReq.sequence_detect | spnego.ContextReq.no_integrity,
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_negotiate)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    token1 = c.step()
    assert isinstance(token1, bytes)

    token2 = s.step(token1)
    assert isinstance(token2, bytes)

    token3 = c.step(token2)
    assert token3 is None

    # Make sure it reports the right protocol
    assert c.negotiated_protocol == "kerberos"
    assert s.negotiated_protocol == "kerberos"

    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr == spnego.ContextReq.mutual_auth | spnego.ContextReq.sequence_detect
    assert s.context_attr == spnego.ContextReq.mutual_auth | spnego.ContextReq.sequence_detect

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)

    token2 = s.step(token1)
    assert isinstance(token2, bytes)

    token3 = c.step(token2)
    assert token3 is None

    # Make sure it reports the right protocol
    assert c.negotiated_protocol == "kerberos"
    assert s.negotiated_protocol == "kerberos"

    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr & spnego.ContextReq.mutual_auth
    assert s.context_attr & spnego.ContextReq.mutual_auth

    _message_test(c, s)


def test_negotiate_through_python_ntlm(ntlm_cred):
    # Build the initial context and assert the defaults.
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        protocol="negotiate",
        context_req=spnego.ContextReq.delegate | spnego.ContextReq.default,
        options=spnego.NegotiateOptions.use_negotiate,
    )
    s = spnego.server(protocol="negotiate")

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    assert not c.complete
    assert not s.complete

    negotiate = c.step()

    assert isinstance(negotiate, bytes)
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)

    assert isinstance(challenge, bytes)
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)

    assert isinstance(authenticate, bytes)
    assert not c.complete
    assert not s.complete

    mech_list_mic = s.step(authenticate)

    assert isinstance(mech_list_mic, bytes)
    assert not c.complete
    assert s.complete

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    mech_list_resp = c.step(mech_list_mic)

    assert mech_list_resp is None
    assert c.complete
    assert s.complete
    assert c.negotiated_protocol == "ntlm"
    assert s.negotiated_protocol == "ntlm"

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    negotiate = c.step()

    assert isinstance(negotiate, bytes)
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)

    assert isinstance(challenge, bytes)
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)

    assert isinstance(authenticate, bytes)
    assert not c.complete
    assert not s.complete

    mech_list_mic = s.step(authenticate)

    assert isinstance(mech_list_mic, bytes)
    assert not c.complete
    assert s.complete

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    mech_list_resp = c.step(mech_list_mic)

    assert mech_list_resp is None
    assert c.complete
    assert s.complete
    assert c.negotiated_protocol == "ntlm"
    assert s.negotiated_protocol == "ntlm"

    _message_test(c, s)


def test_negotiate_with_raw_ntlm(ntlm_cred):
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname(), protocol="ntlm")
    s = spnego.server(options=spnego.NegotiateOptions.use_negotiate)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    negotiate = c.step()
    assert negotiate is not None
    assert negotiate.startswith(b"NTLMSSP\x00\x01")
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)
    assert challenge is not None
    assert challenge.startswith(b"NTLMSSP\x00\x02")
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)
    assert authenticate is not None
    assert authenticate.startswith(b"NTLMSSP\x00\x03")
    assert c.complete
    assert not s.complete

    final = s.step(authenticate)
    assert final is None
    assert c.complete
    assert s.complete

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    negotiate = c.step()
    assert negotiate is not None
    assert negotiate.startswith(b"NTLMSSP\x00\x01")
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)
    assert challenge is not None
    assert challenge.startswith(b"NTLMSSP\x00\x02")
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)
    assert authenticate is not None
    assert authenticate.startswith(b"NTLMSSP\x00\x03")
    assert c.complete
    assert not s.complete

    final = s.step(authenticate)
    assert final is None
    assert c.complete
    assert s.complete

    _message_test(c, s)


def test_negotiate_with_ntlm_hash(ntlm_cred):
    cred = spnego.NTLMHash(username=ntlm_cred[0], nt_hash=ntowfv1(ntlm_cred[1]).hex())
    c = spnego.client(cred, hostname=socket.gethostname())
    s = spnego.server()

    negotiate = c.step()
    assert negotiate is not None
    assert b"NTLMSSP\x00\x01" in negotiate
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)
    assert challenge is not None
    assert b"NTLMSSP\x00\x02" in challenge
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)
    assert authenticate is not None
    assert b"NTLMSSP\x00\x03" in authenticate
    assert not c.complete
    assert not s.complete

    mic = s.step(authenticate)
    assert mic is not None
    assert not c.complete
    assert s.complete

    final = c.step(mic)
    assert final is None
    assert c.complete
    assert s.complete

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    negotiate = c.step()
    assert negotiate is not None
    assert b"NTLMSSP\x00\x01" in negotiate
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)
    assert challenge is not None
    assert b"NTLMSSP\x00\x02" in challenge
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)
    assert authenticate is not None
    assert b"NTLMSSP\x00\x03" in authenticate
    assert not c.complete
    assert not s.complete

    mic = s.step(authenticate)
    assert mic is not None
    assert not c.complete
    assert s.complete

    final = c.step(mic)
    assert final is None
    assert c.complete
    assert s.complete

    _message_test(c, s)


def test_negotiate_with_ntlm_and_duplicate_response_token(ntlm_cred):
    c = spnego.client(
        ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname(), options=spnego.NegotiateOptions.use_negotiate
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_negotiate)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    negotiate = c.step()
    assert not c.complete
    assert not s.complete

    challenge = s.step(negotiate)
    assert challenge is not None

    # Set the mechListMIC to the same value as responseToken in the challenge to replicate the bug this is testing
    neg_token_resp = unpack_token(challenge)
    assert isinstance(neg_token_resp, NegTokenResp)
    neg_token_resp.mech_list_mic = neg_token_resp.response_token
    challenge = neg_token_resp.pack()

    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)
    assert not c.complete
    assert not s.complete

    mech_list = s.step(authenticate)
    assert not c.complete
    assert s.complete

    final = c.step(mech_list)
    assert final is None
    assert c.complete
    assert s.complete

    _message_test(c, s)


# NTLM scenarios


@pytest.mark.parametrize("lm_compat_level", [None, 0, 1, 2])
def test_ntlm_auth(lm_compat_level, ntlm_cred, monkeypatch):
    if lm_compat_level is not None:
        monkeypatch.setenv("LM_COMPAT_LEVEL", str(lm_compat_level))

    # Build the initial context and assert the defaults.
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], protocol="ntlm", options=spnego.NegotiateOptions.use_ntlm)
    s = spnego.server(protocol="ntlm", options=spnego.NegotiateOptions.use_ntlm)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    with pytest.warns(DeprecationWarning, match="username is deprecated"):
        c.username

    with pytest.warns(DeprecationWarning, match="password is deprecated"):
        c.password

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)


@pytest.mark.parametrize(
    "client_opt, server_opt",
    [
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_sspi),
        (spnego.NegotiateOptions.use_ntlm, spnego.NegotiateOptions.use_sspi),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_ntlm),
        (spnego.NegotiateOptions.use_ntlm, spnego.NegotiateOptions.use_ntlm),
        # Cannot test with gssapi as the existing version has a bug with this scenario.
    ],
)
def test_sspi_ntlm_auth_no_sign_or_seal(client_opt, server_opt, ntlm_cred):
    if client_opt & spnego.NegotiateOptions.use_gssapi or server_opt & spnego.NegotiateOptions.use_gssapi:
        if "ntlm" not in spnego._gss.GSSAPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through GSSAPI")

    elif client_opt & spnego.NegotiateOptions.use_sspi or server_opt & spnego.NegotiateOptions.use_sspi:
        if "ntlm" not in spnego._sspi.SSPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through SSPI")

    # Build the initial context and assert the defaults.
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=client_opt,
        protocol="ntlm",
        context_req=spnego.ContextReq.none,
    )
    s = spnego.server(options=server_opt, protocol="ntlm", context_req=spnego.ContextReq.none)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    # Client sign, server verify
    plaintext = os.urandom(3)

    c_sig = c.sign(plaintext)
    s.verify(plaintext, c_sig)

    # Server sign, client verify
    plaintext = os.urandom(9)

    s_sig = s.sign(plaintext)
    c.verify(plaintext, s_sig)


@pytest.mark.skipif(
    "ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Test requires NTLM to be available through SSPI"
)
@pytest.mark.parametrize(
    "client_opt, server_opt, cbt, cbt_with_step",
    [
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_sspi, False, False),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_sspi, True, False),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_sspi, True, True),
        (spnego.NegotiateOptions.use_ntlm, spnego.NegotiateOptions.use_sspi, False, False),
        (spnego.NegotiateOptions.use_ntlm, spnego.NegotiateOptions.use_sspi, True, False),
        (spnego.NegotiateOptions.use_ntlm, spnego.NegotiateOptions.use_sspi, True, True),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_ntlm, False, False),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_ntlm, True, False),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_ntlm, True, False),
        (spnego.NegotiateOptions.use_sspi, spnego.NegotiateOptions.use_ntlm, True, True),
    ],
)
def test_sspi_ntlm_auth(client_opt, server_opt, cbt, cbt_with_step, ntlm_cred):
    # Build the initial context and assert the defaults.
    step_kwargs: typing.Dict[str, typing.Any] = {}
    client_kwargs: typing.Dict[str, typing.Any] = {}
    server_kwargs: typing.Dict[str, typing.Any] = {}

    if cbt:
        bindings = spnego.channel_bindings.GssChannelBindings(application_data=b"test_data:\x00\x01")
        if cbt_with_step:
            step_kwargs["channel_bindings"] = bindings
            client_kwargs["channel_bindings"] = spnego.channel_bindings.GssChannelBindings(
                application_data=b"test_data:\x00\x02"
            )
            server_kwargs["channel_bindings"] = spnego.channel_bindings.GssChannelBindings(
                application_data=b"test_data:\x00\x03"
            )
        else:
            client_kwargs["channel_bindings"] = bindings
            server_kwargs["channel_bindings"] = bindings

    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=client_opt,
        protocol="ntlm",
        **client_kwargs,
    )
    s = spnego.server(options=server_opt, protocol="ntlm", **server_kwargs)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    _ntlm_test(c, s, **step_kwargs)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    _ntlm_test(c, s, **step_kwargs)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)


@pytest.mark.skipif(
    "ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Test requires NTLM to be available through SSPI"
)
@pytest.mark.parametrize("lm_compat_level", [1, 2, 3])
def test_sspi_ntlm_lm_compat(lm_compat_level, ntlm_cred, monkeypatch):
    monkeypatch.setenv("LM_COMPAT_LEVEL", str(lm_compat_level))
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        protocol="ntlm",
        options=spnego.NegotiateOptions.use_ntlm,
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_sspi, protocol="ntlm")

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)


def test_ntlm_with_explicit_ntlm_hash(ntlm_cred):
    ntlm_hashes = f"{lmowfv1(ntlm_cred[1]).hex()}:{ntowfv1(ntlm_cred[1]).hex()}"
    c = spnego.client(
        ntlm_cred[0], ntlm_hashes, hostname=socket.gethostname(), options=spnego.NegotiateOptions.none, protocol="ntlm"
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()
    _ntlm_test(c, s)

    assert c.client_principal is None
    assert s.client_principal == ntlm_cred[0]

    _message_test(c, s)


def test_ntlm_with_unsupported_credential():
    with pytest.raises(
        InvalidCredentialError, match="Invalid username/credential specified, must be a string or Credential object."
    ):
        spnego.client(123, protocol="ntlm")  # type: ignore[arg-type] # Testing this scenario


# Kerberos scenarios


@pytest.mark.parametrize("explicit_user", [False, True])
def test_gssapi_kerberos_auth(explicit_user, kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    username = None
    if explicit_user:
        username = kerb_cred.user_princ

    c = spnego.client(
        username, None, hostname=kerb_cred.hostname, protocol="kerberos", options=spnego.NegotiateOptions.use_gssapi
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_gssapi, protocol="kerberos")

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    with pytest.raises(SpnegoError, match="Retrieving session key"):
        _ = c.session_key

    with pytest.raises(SpnegoError, match="Retrieving session key"):
        _ = s.session_key

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)


@pytest.mark.parametrize("explicit_user", [False, True])
def test_gssapi_kerberos_auth_no_integrity(explicit_user, kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    username = None
    if explicit_user:
        username = kerb_cred.user_princ

    c = spnego.client(
        username,
        None,
        hostname=kerb_cred.hostname,
        protocol="kerberos",
        options=spnego.NegotiateOptions.use_gssapi,
        context_req=spnego.ContextReq.mutual_auth | spnego.ContextReq.sequence_detect | spnego.ContextReq.no_integrity,
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_gssapi, protocol="kerberos")

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    with pytest.raises(SpnegoError, match="Retrieving session key"):
        _ = c.session_key

    with pytest.raises(SpnegoError, match="Retrieving session key"):
        _ = s.session_key

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr == spnego.ContextReq.mutual_auth | spnego.ContextReq.sequence_detect
    assert s.context_attr == spnego.ContextReq.mutual_auth | spnego.ContextReq.sequence_detect

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    print(c.context_attr)
    print(s.context_attr)


@pytest.mark.parametrize("acquire_cred_from", [False, True])
def test_gssapi_kerberos_auth_explicit_cred(acquire_cred_from, kerb_cred, monkeypatch):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    if not acquire_cred_from:
        monkeypatch.delattr("gssapi.raw.acquire_cred_from", raising=False)

    context_req = spnego.ContextReq.default | spnego.ContextReq.delegate
    c = spnego.client(
        kerb_cred.user_princ,
        kerb_cred.password("user"),
        hostname=socket.getfqdn(),
        protocol="kerberos",
        options=spnego.NegotiateOptions.use_gssapi,
        context_req=context_req,
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_gssapi, protocol="kerberos")

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"

    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    with pytest.raises(SpnegoError, match="Retrieving session key"):
        _ = c.session_key

    with pytest.raises(SpnegoError, match="Retrieving session key"):
        _ = s.session_key

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr & spnego.ContextReq.delegate
    assert s.context_attr & spnego.ContextReq.delegate

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    assert c.context_attr & spnego.ContextReq.delegate
    assert s.context_attr & spnego.ContextReq.delegate

    _message_test(c, s)


@pytest.mark.parametrize(
    "protocol, set_principal",
    [
        ("kerberos", False),
        ("kerberos", True),
        ("negotiate", False),
        ("negotiate", True),
    ],
)
def test_kerberos_auth_keytab(protocol, set_principal, kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    kerb_cred.extract_keytab(kerb_cred.user_princ, kerb_cred.client_keytab)
    if set_principal:
        kt = spnego.KerberosKeytab(keytab=kerb_cred.client_keytab, principal=kerb_cred.user_princ)
    else:
        kt = spnego.KerberosKeytab(keytab=kerb_cred.client_keytab)

    context_req = spnego.ContextReq.default
    c = spnego.client(kt, hostname=socket.getfqdn(), protocol=protocol, context_req=context_req)
    s = spnego.server(protocol=protocol)

    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)


@pytest.mark.parametrize(
    "protocol, explicit_user",
    [
        ("kerberos", False),
        ("kerberos", True),
        ("negotiate", False),
        ("kerberos", True),
    ],
)
def test_kerberos_auth_ccache(protocol, explicit_user, kerb_cred, monkeypatch):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    # Verified we are actually using our explicit CCache
    monkeypatch.setenv("KRB5CCNAME", "missing")

    context_req = spnego.ContextReq.default

    if explicit_user:
        ccache = spnego.KerberosCCache(ccache=kerb_cred.ccache, principal=kerb_cred.user_princ)
    else:
        ccache = spnego.KerberosCCache(ccache=kerb_cred.ccache)

    c = spnego.client(ccache, hostname=socket.getfqdn(), protocol=protocol, context_req=context_req)
    s = spnego.server(protocol=protocol)

    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"
    # assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)


@pytest.mark.parametrize(
    "protocol, explicit_user",
    [
        ("kerberos", False),
        ("kerberos", True),
        ("negotiate", False),
        ("kerberos", True),
    ],
)
def test_kerberos_auth_env_cache(protocol, explicit_user, kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    context_req = spnego.ContextReq.default
    cred = None
    if explicit_user:
        cred = spnego.CredentialCache(username=kerb_cred.user_princ)

    c = spnego.client(cred, hostname=socket.getfqdn(), protocol=protocol, context_req=context_req)
    s = spnego.server(protocol=protocol)

    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)


@pytest.mark.parametrize(
    "protocol, with_step",
    [
        ("kerberos", False),
        ("kerberos", True),
        ("negotiate", False),
        ("negotiate", True),
    ],
)
def test_kerberos_auth_channel_bindings(protocol, with_step, kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    cbt = spnego.channel_bindings.GssChannelBindings(application_data=b"test_data:\x00\x01")

    step_kwargs: typing.Dict[str, typing.Any] = {}
    client_kwargs: typing.Dict[str, typing.Any] = {}
    server_kwargs: typing.Dict[str, typing.Any] = {}
    if with_step:
        step_kwargs["channel_bindings"] = cbt

        # Use a fake value to ensure step takes priority.
        client_kwargs["channel_bindings"] = spnego.channel_bindings.GssChannelBindings(
            application_data=b"test_data:\x00\x02"
        )
        server_kwargs["channel_bindings"] = spnego.channel_bindings.GssChannelBindings(
            application_data=b"test_data:\x00\x03"
        )
    else:
        client_kwargs["channel_bindings"] = cbt
        server_kwargs["channel_bindings"] = cbt

    c = spnego.client(
        kerb_cred.user_princ,
        kerb_cred.password("user"),
        hostname=socket.getfqdn(),
        protocol=protocol,
        **client_kwargs,
    )
    s = spnego.server(protocol=protocol, **server_kwargs)

    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token1 = c.step(**step_kwargs)
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"
    # assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1, **step_kwargs)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2, **step_kwargs)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    token1 = c.step(**step_kwargs)
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    if protocol == "negotiate":
        assert s.negotiated_protocol is None
    else:
        assert s.negotiated_protocol == "kerberos"

    token2 = s.step(token1, **step_kwargs)
    assert isinstance(token2, bytes)
    assert not c.complete
    assert s.complete
    assert s.negotiated_protocol == "kerberos"

    token3 = c.step(token2, **step_kwargs)
    assert token3 is None
    assert c.complete
    assert s.complete
    assert isinstance(c.session_key, bytes)
    assert isinstance(s.session_key, bytes)
    assert c.session_key == s.session_key

    assert c.client_principal is None
    assert s.client_principal == kerb_cred.user_princ

    _message_test(c, s)


# CredSSP scenarios


@pytest.mark.parametrize(
    "options, restrict_tlsv12, version",
    [
        (spnego.NegotiateOptions.use_negotiate, False, None),
        (spnego.NegotiateOptions.use_negotiate, False, 2),
        (spnego.NegotiateOptions.use_negotiate, True, None),
        # Using NTLM directly results in a slightly separate behaviour for the pub key.
        (spnego.NegotiateOptions.use_ntlm, False, None),
        (spnego.NegotiateOptions.use_ntlm, False, 5),
        (spnego.NegotiateOptions.use_ntlm, True, None),
    ],
)
def test_credssp_ntlm_creds(options, restrict_tlsv12, version, ntlm_cred, monkeypatch, tmp_path):
    context_kwargs: typing.Dict[str, typing.Any] = {}
    if restrict_tlsv12:
        credssp_context = spnego.tls.default_tls_context(usage="accept")

        tls_version = getattr(ssl, "TLSVersion", None)
        if hasattr(credssp_context.context, "maximum_version") and tls_version:
            setattr(credssp_context.context, "maximum_version", tls_version.TLSv1_2)

        else:
            credssp_context.context.options |= ssl.Options.OP_NO_TLSv1_3

        cert_pem, key_pem, pub_key = spnego.tls.generate_tls_certificate()
        with open(tmp_path / "ca.pem", mode="wb") as fd:
            fd.write(cert_pem)
            fd.write(key_pem)

        credssp_context.context.load_cert_chain(tmp_path / "ca.pem")
        credssp_context.public_key = pub_key
        context_kwargs["credssp_tls_context"] = credssp_context

    if version:
        monkeypatch.setattr(spnego._credssp, "_CREDSSP_VERSION", version)
    else:
        version = spnego._credssp._CREDSSP_VERSION

    c = spnego.client(ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname(), protocol="credssp", options=options)
    s = spnego.server(protocol="credssp", options=options, **context_kwargs)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"
    assert c.get_extra_info("client_credential") is None
    assert isinstance(c.get_extra_info("sslcontext"), ssl.SSLContext)
    assert isinstance(c.get_extra_info("ssl_object"), ssl.SSLObject)
    assert c.get_extra_info("auth_stage") == "TLS Handshake"

    assert s.get_extra_info("client_credential") is None
    assert isinstance(s.get_extra_info("sslcontext"), ssl.SSLContext)
    assert isinstance(s.get_extra_info("ssl_object"), ssl.SSLObject)
    assert s.get_extra_info("auth_stage") == "TLS Handshake"

    assert c.client_principal is None
    assert c.get_extra_info("client_credential") is None
    assert c.negotiated_protocol is None

    server_tls_token = None
    while c.get_extra_info("auth_stage") == "TLS Handshake":
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)
    assert c.get_extra_info("auth_stage") == "Public key exchange"
    assert s.get_extra_info("auth_stage").startswith("Authentication")
    assert not c.complete
    assert not s.complete

    server_pub_key = s.step(ntlm3_pub_key)
    assert c.get_extra_info("auth_stage") == "Public key exchange"
    assert s.get_extra_info("auth_stage") == "Public key exchange"
    assert not c.complete
    assert not s.complete

    credential = c.step(server_pub_key)
    assert c.get_extra_info("auth_stage") == "Credential exchange"
    assert s.get_extra_info("auth_stage") == "Public key exchange"
    assert c.complete
    assert not s.complete

    final_token = s.step(credential)
    assert c.get_extra_info("auth_stage") == "Credential exchange"
    assert s.get_extra_info("auth_stage") == "Credential exchange"
    assert final_token is None
    assert c.complete
    assert s.complete

    assert c.negotiated_protocol == "ntlm"
    assert s.negotiated_protocol == "ntlm"

    domain, username = ntlm_cred[0].split("\\")

    assert c.get_extra_info("client_credential") is None
    client_credential = s.get_extra_info("client_credential")
    assert isinstance(client_credential, TSPasswordCreds)
    assert client_credential.username == username
    assert client_credential.domain_name == domain
    assert client_credential.password == ntlm_cred[1]

    assert c.get_extra_info("protocol_version") == version
    assert s.get_extra_info("protocol_version") == version

    assert isinstance(c.query_message_sizes().header, int)
    assert isinstance(s.query_message_sizes().header, int)

    _message_test(c, s)

    plaintext = os.urandom(16)
    c_winrm_result = c.wrap_winrm(plaintext)
    assert isinstance(c_winrm_result, WinRMWrapResult)
    assert isinstance(c_winrm_result.header, bytes)
    assert isinstance(c_winrm_result.data, bytes)
    assert isinstance(c_winrm_result.padding_length, int)

    s_winrm_result = s.unwrap_winrm(c_winrm_result.header, c_winrm_result.data)
    assert s_winrm_result == plaintext

    plaintext = os.urandom(16)
    s_winrm_result = s.wrap_winrm(plaintext)
    assert isinstance(s_winrm_result, WinRMWrapResult)
    assert isinstance(s_winrm_result.header, bytes)
    assert isinstance(s_winrm_result.data, bytes)
    assert isinstance(s_winrm_result.padding_length, int)

    c_winrm_result = c.unwrap_winrm(s_winrm_result.header, s_winrm_result.data)
    assert c_winrm_result == plaintext

    c = c.new_context()
    s = s.new_context()

    server_tls_token = None
    while c.get_extra_info("auth_stage") == "TLS Handshake":
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)
    assert c.get_extra_info("auth_stage") == "Public key exchange"
    assert s.get_extra_info("auth_stage").startswith("Authentication")
    assert not c.complete
    assert not s.complete

    server_pub_key = s.step(ntlm3_pub_key)
    assert c.get_extra_info("auth_stage") == "Public key exchange"
    assert s.get_extra_info("auth_stage") == "Public key exchange"
    assert not c.complete
    assert not s.complete

    credential = c.step(server_pub_key)
    assert c.get_extra_info("auth_stage") == "Credential exchange"
    assert s.get_extra_info("auth_stage") == "Public key exchange"
    assert c.complete
    assert not s.complete

    final_token = s.step(credential)
    assert c.get_extra_info("auth_stage") == "Credential exchange"
    assert s.get_extra_info("auth_stage") == "Credential exchange"
    assert final_token is None
    assert c.complete
    assert s.complete

    assert c.negotiated_protocol == "ntlm"
    assert s.negotiated_protocol == "ntlm"

    domain, username = ntlm_cred[0].split("\\")

    assert c.get_extra_info("client_credential") is None
    client_credential = s.get_extra_info("client_credential")
    assert isinstance(client_credential, TSPasswordCreds)
    assert client_credential.username == username
    assert client_credential.domain_name == domain
    assert client_credential.password == ntlm_cred[1]

    assert c.get_extra_info("protocol_version") == version
    assert s.get_extra_info("protocol_version") == version

    _message_test(c, s)

    plaintext = os.urandom(16)
    c_winrm_result = c.wrap_winrm(plaintext)
    assert isinstance(c_winrm_result, WinRMWrapResult)
    assert isinstance(c_winrm_result.header, bytes)
    assert isinstance(c_winrm_result.data, bytes)
    assert isinstance(c_winrm_result.padding_length, int)

    s_winrm_result = s.unwrap_winrm(c_winrm_result.header, c_winrm_result.data)
    assert s_winrm_result == plaintext

    plaintext = os.urandom(16)
    s_winrm_result = s.wrap_winrm(plaintext)
    assert isinstance(s_winrm_result, WinRMWrapResult)
    assert isinstance(s_winrm_result.header, bytes)
    assert isinstance(s_winrm_result.data, bytes)
    assert isinstance(s_winrm_result.padding_length, int)

    c_winrm_result = c.unwrap_winrm(s_winrm_result.header, s_winrm_result.data)
    assert c_winrm_result == plaintext


@pytest.mark.parametrize("restrict_tlsv12", [False, True])
def test_credssp_kerberos_creds(restrict_tlsv12, kerb_cred):
    if sys.platform == "darwin":
        pytest.skip("Environment problem with GSS.Framework - skip")

    c_kerb_context = spnego.client(kerb_cred.user_princ, None, hostname=socket.getfqdn(), protocol="kerberos")
    s_kerb_context = spnego.server(protocol="kerberos")

    client_kwargs: typing.Dict[str, typing.Any] = {}
    if restrict_tlsv12:
        tls_context = spnego.tls.default_tls_context()

        tls_version = getattr(ssl, "TLSVersion", None)
        if hasattr(tls_context.context, "maximum_version") and tls_version:
            setattr(tls_context.context, "maximum_version", tls_version.TLSv1_2)

        else:
            tls_context.context.options |= ssl.Options.OP_NO_TLSv1_3

        client_kwargs["credssp_tls_context"] = tls_context

    c = spnego.client(
        kerb_cred.user_princ,
        kerb_cred.password("user"),
        protocol="credssp",
        credssp_negotiate_context=c_kerb_context,
        **client_kwargs,
    )
    s = spnego.server(protocol="credssp", credssp_negotiate_context=s_kerb_context)

    assert c.get_extra_info("invalid") is None
    assert c.get_extra_info("invalid", "default") == "default"
    assert c.get_extra_info("client_credential") is None
    assert isinstance(c.get_extra_info("sslcontext"), ssl.SSLContext)
    assert isinstance(c.get_extra_info("ssl_object"), ssl.SSLObject)

    assert s.get_extra_info("client_credential") is None
    assert isinstance(s.get_extra_info("sslcontext"), ssl.SSLContext)
    assert isinstance(s.get_extra_info("ssl_object"), ssl.SSLObject)

    server_token = None
    while c.get_extra_info("auth_stage") != "Public key exchange":
        client_token = c.step(server_token)
        assert not c.complete
        assert not s.complete

        server_token = s.step(client_token)
        assert not c.complete
        assert not s.complete

    assert s_kerb_context.complete

    credential = c.step(server_token)
    assert c.complete
    assert not s.complete

    final_token = s.step(credential)
    assert final_token is None
    assert c.complete
    assert s.complete

    assert c.negotiated_protocol == "kerberos"
    assert s.negotiated_protocol == "kerberos"
    assert c_kerb_context.negotiated_protocol == "kerberos"
    assert s_kerb_context.negotiated_protocol == "kerberos"

    assert s.client_principal == kerb_cred.user_princ
    assert c.get_extra_info("client_credential") is None
    client_credential = s.get_extra_info("client_credential")
    assert isinstance(client_credential, TSPasswordCreds)
    assert client_credential.username == kerb_cred.user_princ
    assert client_credential.password == kerb_cred.password("user")

    _message_test(c, s)

    plaintext = os.urandom(16)
    c_winrm_result = c.wrap_winrm(plaintext)
    assert isinstance(c_winrm_result, WinRMWrapResult)
    assert isinstance(c_winrm_result.header, bytes)
    assert isinstance(c_winrm_result.data, bytes)
    assert isinstance(c_winrm_result.padding_length, int)

    s_winrm_result = s.unwrap_winrm(c_winrm_result.header, c_winrm_result.data)
    assert s_winrm_result == plaintext

    plaintext = os.urandom(16)
    s_winrm_result = s.wrap_winrm(plaintext)
    assert isinstance(s_winrm_result, WinRMWrapResult)
    assert isinstance(s_winrm_result.header, bytes)
    assert isinstance(s_winrm_result.data, bytes)
    assert isinstance(s_winrm_result.padding_length, int)

    c_winrm_result = c.unwrap_winrm(s_winrm_result.header, s_winrm_result.data)
    assert c_winrm_result == plaintext

    c = c.new_context()
    s = s.new_context()

    server_token = None
    while c.get_extra_info("auth_stage") != "Public key exchange":
        client_token = c.step(server_token)
        assert not c.complete
        assert not s.complete

        server_token = s.step(client_token)
        assert not c.complete
        assert not s.complete

    assert s_kerb_context.complete

    credential = c.step(server_token)
    assert c.complete
    assert not s.complete

    final_token = s.step(credential)
    assert final_token is None
    assert c.complete
    assert s.complete

    assert c.negotiated_protocol == "kerberos"
    assert s.negotiated_protocol == "kerberos"
    assert c_kerb_context.negotiated_protocol == "kerberos"
    assert s_kerb_context.negotiated_protocol == "kerberos"

    assert s.client_principal == kerb_cred.user_princ
    assert c.get_extra_info("client_credential") is None
    client_credential = s.get_extra_info("client_credential")
    assert isinstance(client_credential, TSPasswordCreds)
    assert client_credential.username == kerb_cred.user_princ
    assert client_credential.password == kerb_cred.password("user")

    _message_test(c, s)

    plaintext = os.urandom(16)
    c_winrm_result = c.wrap_winrm(plaintext)
    assert isinstance(c_winrm_result, WinRMWrapResult)
    assert isinstance(c_winrm_result.header, bytes)
    assert isinstance(c_winrm_result.data, bytes)
    assert isinstance(c_winrm_result.padding_length, int)

    s_winrm_result = s.unwrap_winrm(c_winrm_result.header, c_winrm_result.data)
    assert s_winrm_result == plaintext

    plaintext = os.urandom(16)
    s_winrm_result = s.wrap_winrm(plaintext)
    assert isinstance(s_winrm_result, WinRMWrapResult)
    assert isinstance(s_winrm_result.header, bytes)
    assert isinstance(s_winrm_result.data, bytes)
    assert isinstance(s_winrm_result.padding_length, int)

    c_winrm_result = c.unwrap_winrm(s_winrm_result.header, s_winrm_result.data)
    assert c_winrm_result == plaintext


def test_credssp_multiple_creds(ntlm_cred):
    creds: typing.List[spnego.Credential] = [
        spnego.NTLMHash(username=ntlm_cred[0], nt_hash=ntowfv1(ntlm_cred[1]).hex()),
        spnego.Password(username="delegate", password="password"),
    ]
    c = spnego.client(creds, hostname=socket.gethostname(), protocol="credssp")
    s = spnego.server(protocol="credssp")

    assert c.client_principal is None
    assert c.negotiated_protocol is None
    assert c.get_extra_info("client_credential") is None

    server_tls_token = None
    while c.get_extra_info("auth_stage") == "TLS Handshake":
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)
    assert not c.complete
    assert not s.complete

    server_pub_key = s.step(ntlm3_pub_key)
    assert not c.complete
    assert not s.complete

    credential = c.step(server_pub_key)
    assert c.complete
    assert not s.complete

    final_token = s.step(credential)
    assert final_token is None
    assert c.complete
    assert s.complete

    assert c.negotiated_protocol == "ntlm"
    assert s.negotiated_protocol == "ntlm"

    # Matches the principal authenticated with the Negotiate phase - NTLMHash
    assert s.client_principal == ntlm_cred[0]

    # Matches the Password cred passed in, will not use the NTLMHash details as it wasn't enought for CredSSP
    client_credential = s.get_extra_info("client_credential")
    assert isinstance(client_credential, TSPasswordCreds)
    assert client_credential.username == "delegate"
    assert client_credential.domain_name == ""
    assert client_credential.password == "password"

    _message_test(c, s)

    c = c.new_context()
    s = s.new_context()

    assert c.client_principal is None
    assert c.negotiated_protocol is None
    assert c.get_extra_info("client_credential") is None

    server_tls_token = None
    while c.get_extra_info("auth_stage") == "TLS Handshake":
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)
    assert not c.complete
    assert not s.complete

    server_pub_key = s.step(ntlm3_pub_key)
    assert not c.complete
    assert not s.complete

    credential = c.step(server_pub_key)
    assert c.complete
    assert not s.complete

    final_token = s.step(credential)
    assert final_token is None
    assert c.complete
    assert s.complete

    assert c.negotiated_protocol == "ntlm"
    assert s.negotiated_protocol == "ntlm"

    # Matches the principal authenticated with the Negotiate phase - NTLMHash
    assert s.client_principal == ntlm_cred[0]

    # Matches the Password cred passed in, will not use the NTLMHash details as it wasn't enought for CredSSP
    client_credential = s.get_extra_info("client_credential")
    assert isinstance(client_credential, TSPasswordCreds)
    assert client_credential.username == "delegate"
    assert client_credential.domain_name == ""
    assert client_credential.password == "password"

    _message_test(c, s)


def test_credssp_min_protocol_failure_initiator(ntlm_cred, monkeypatch):
    monkeypatch.setattr(spnego._credssp, "_CREDSSP_VERSION", 4)

    c = spnego.client(
        ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname(), protocol="credssp", credssp_min_protocol=5
    )
    s = spnego.server(protocol="credssp")

    # The TLS handshake can differ based on the protocol selected, keep on looping until we see the auth_context set up
    # For NTLM the auth context will be present after the first exchange of NTLM tokens.
    server_tls_token = None
    while c._auth_context is None:  # type: ignore[attr-defined]
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    with pytest.raises(
        InvalidTokenError, match="The peer protocol version was 4 and did not meet the minimum requirements of 5"
    ):
        c.step(server_tls_token)


def test_credssp_min_protocol_failure_acceptor(ntlm_cred, monkeypatch):
    monkeypatch.setattr(spnego._credssp, "_CREDSSP_VERSION", 4)

    c = spnego.client(ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname(), protocol="credssp")
    s = spnego.server(protocol="credssp", credssp_min_protocol=5)

    # The TLS handshake can differ based on the protocol selected, keep on looping until we see the auth_context set up
    # For NTLM the auth context will be present after the first exchange of NTLM tokens.
    server_tls_token = None
    while True:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        if c.get_extra_info("auth_stage").startswith("Authentication"):
            break

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    with pytest.raises(
        InvalidTokenError, match="The peer protocol version was 4 and did not meet the minimum requirements of 5"
    ):
        s.step(client_tls_token)



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)
import typing

import pytest

import spnego._text as text


class Obj:
    def __init__(self, str: typing.Optional[str] = None, repr: typing.Optional[str] = None) -> None:
        self.str = str
        self.repr = repr

    def __eq__(self, other: typing.Any) -> bool:
        return isinstance(other, Obj) and self.str == other.str and self.repr == other.repr

    def __str__(self) -> str:
        return self.str  # type: ignore[return-value] # Testing this scenario

    def __repr__(self) -> str:
        return self.repr  # type: ignore[return-value] # Testing this scenario


class ObjRepr:
    def __init__(self, repr: typing.Optional[str] = None) -> None:
        self.repr = repr

    def __repr__(self) -> str:
        return self.repr  # type: ignore[return-value] # Testing this scenario


class ObjUnicodeError:
    def __str__(self) -> str:
        return "café".encode("ascii")  # type: ignore[return-value] # Testing this scenario

    def __repr__(self) -> str:
        return "café".encode("ascii")  # type: ignore[return-value] # Testing this scenario


def test_to_bytes_from_bytes():
    actual = text.to_bytes(b"\xFF\x00\x7F\x80")

    assert actual == b"\xFF\x00\x7F\x80"


@pytest.mark.parametrize(
    "value, expected",
    [
        ("cafe", b"cafe"),
        ("café", b"caf\xc3\xa9"),
        ("ÜseӜ", b"\x55\xCC\x88\x73\x65\xD3\x9C"),
    ],
)
def test_to_bytes_from_text(value, expected):
    actual = text.to_bytes(value)

    assert actual == expected


def test_to_bytes_encoding():
    actual = text.to_bytes("café", encoding="windows-1252")

    assert actual == b"caf\xe9"


def test_to_bytes_errors():
    with pytest.raises(UnicodeEncodeError, match="codec can't encode character"):
        text.to_bytes("café", encoding="ascii")

    actual = text.to_bytes("café", encoding="ascii", errors="replace")

    assert actual == b"caf?"


def test_to_bytes_nonstr():
    actual = text.to_bytes(Obj(str="café"))

    assert actual == b"caf\xc3\xa9"


def test_to_bytes_nonstr_default():
    actual = text.to_bytes(Obj())

    assert actual == b""


def test_to_bytes_nonstr_repr():
    actual = text.to_bytes(ObjRepr(repr="café"))

    assert actual == b"caf\xc3\xa9"


def test_to_bytes_nonstr_unicode_error():
    actual = text.to_bytes(ObjUnicodeError())

    assert actual == b""


def test_to_bytes_nonstr_passthru():
    actual = text.to_bytes(Obj(), nonstring="passthru")

    assert actual == Obj()


def test_to_bytes_nonstr_empty():
    actual = text.to_bytes(Obj(), nonstring="empty")

    assert actual == b""


def test_to_bytes_nonstr_invalid():
    with pytest.raises(ValueError, match="Invalid nonstring value"):
        text.to_bytes(Obj(), nonstring="invalid")


def test_to_text_from_text():
    actual = text.to_text("café")

    assert actual == "café"


@pytest.mark.parametrize(
    "value, expected",
    [
        (b"cafe", "cafe"),
        (b"caf\xc3\xa9", "café"),
        (b"\x55\xCC\x88\x73\x65\xD3\x9C", "ÜseӜ"),
    ],
)
def test_to_text_from_bytes(value, expected):
    actual = text.to_text(value)

    assert actual == expected


def test_to_text_encoding():
    actual = text.to_text(b"caf\xe9", encoding="windows-1252")

    assert actual == "café"


def test_to_text_errors():
    with pytest.raises(UnicodeError, match="codec can't decode byte 0xff"):
        text.to_text(b"caf\xFF")

    actual = text.to_text(b"caf\xFF", errors="replace")

    assert actual == "caf�"


def test_to_text_nonstr():
    actual = text.to_text(Obj(str="café"))

    assert actual == "café"


def test_to_text_nonstr_unicode():
    class ObjUnicode:
        def __unicode__(self):
            return "café"

    actual = text.to_text(ObjUnicode())

    assert actual == "café"


def test_to_text_nonstr_default():
    actual = text.to_text(Obj())

    assert actual == ""


def test_to_text_nonstr_repr():
    actual = text.to_text(ObjRepr(repr="café"))

    assert actual == "café"


def test_to_text_nonstr_unicode_error():
    actual = text.to_text(ObjUnicodeError())

    assert actual == ""


def test_to_text_nonstr_passthru():
    actual = text.to_text(Obj(), nonstring="passthru")

    assert actual == Obj()


def test_to_text_nonstr_empty():
    actual = text.to_text(Obj(), nonstring="empty")

    assert actual == ""


def test_to_text_nonstr_invalid():
    with pytest.raises(ValueError, match="Invalid nonstring value"):
        text.to_text(Obj(), nonstring="invalid")



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import io
import json
import os.path

import pytest

import spnego.__main__ as entrypoint
from spnego._ntlm_raw.messages import (
    Authenticate,
    AvId,
    Challenge,
    Negotiate,
    SingleHost,
    TargetInfo,
)
from spnego._spnego import NegTokenResp

from .conftest import get_data


def test_parse_with_hex(capsys):
    entrypoint.main(["--token", base64.b16encode(get_data("ntlm_negotiate")).decode().lower()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None
    assert actual["Data"] is not None


def test_parse_with_file(capsys):
    entrypoint.main(["--file", os.path.join(os.path.dirname(__file__), "data", "ntlm_negotiate")])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None
    assert actual["Data"] is not None


def test_parse_with_file_missing():
    with pytest.raises(ValueError, match="Cannot find file at path"):
        entrypoint.main(["--file", "missing"])


def test_parse_from_stdin(capsys, monkeypatch):
    monkeypatch.setattr("sys.stdin", io.TextIOWrapper(io.BytesIO(get_data("ntlm_negotiate"))))
    entrypoint.main([])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None
    assert actual["Data"] is not None


def test_parse_from_stdin_base16(capsys, monkeypatch):
    monkeypatch.setattr("sys.stdin", io.TextIOWrapper(io.BytesIO(base64.b16encode(get_data("ntlm_negotiate")))))
    entrypoint.main([])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None
    assert actual["Data"] is not None


def test_parse_from_stdin_base64(capsys, monkeypatch):
    monkeypatch.setattr("sys.stdin", io.TextIOWrapper(io.BytesIO(base64.b64encode(get_data("ntlm_negotiate")))))
    entrypoint.main([])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None
    assert actual["Data"] is not None


def test_parse_output_yaml(capsys):
    ruamel = pytest.importorskip("ruamel")
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("ntlm_negotiate")).decode(),
            "--format",
            "yaml",
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    with pytest.raises(ValueError):
        json.loads(actual_out.out)

    loader = ruamel.yaml.YAML(typ="safe", pure=True)
    actual = loader.load(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None
    assert actual["Data"] is not None


def test_parse_output_yaml_not_installed(monkeypatch):
    monkeypatch.setattr(entrypoint, "HAS_YAML", False)

    with pytest.raises(ValueError, match="Cannot output as yaml as ruamel.yaml is not installed"):
        entrypoint.main(
            [
                "--token",
                base64.b64encode(get_data("ntlm_negotiate")).decode(),
                "--format",
                "yaml",
            ]
        )


def test_parse_invalid_spnego_token(capsys):
    resp = NegTokenResp(response_token=b"invalid")
    entrypoint.main(
        [
            "--token",
            base64.b64encode(resp.pack()).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "SPNEGO NegTokenResp"
    assert actual["RawData"] is not None
    assert actual["Data"]["negState"] is None
    assert actual["Data"]["supportedMech"] is None
    assert actual["Data"]["responseToken"]["MessageType"] == "Unknown - Failed to parse see Data for more details."
    assert (
        actual["Data"]["responseToken"]["Data"]
        == "Failed to parse token: Expecting a tag number of 0 not 9 for InitialContextToken"
    )
    assert actual["Data"]["responseToken"]["RawData"] == "696E76616C6964"
    assert actual["Data"]["mechListMIC"] is None


def test_ntlm_negotiate(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("ntlm_negotiate")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "NEGOTIATE_MESSAGE (1)"
    assert actual["RawData"] is not None

    assert actual["Data"]["NegotiateFlags"]["raw"] == 3792208567
    assert set(actual["Data"]["NegotiateFlags"]["flags"]) == {
        "NTLMSSP_NEGOTIATE_56 (2147483648)",
        "NTLMSSP_NEGOTIATE_KEY_EXCH (1073741824)",
        "NTLMSSP_NEGOTIATE_128 (536870912)",
        "NTLMSSP_NEGOTIATE_VERSION (33554432)",
        "NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY (524288)",
        "NTLMSSP_NEGOTIATE_ALWAYS_SIGN (32768)",
        "NTLMSSP_NEGOTIATE_NTLM (512)",
        "NTLMSSP_NEGOTIATE_LM_KEY (128)",
        "NTLMSSP_NEGOTIATE_SEAL (32)",
        "NTLMSSP_NEGOTIATE_SIGN (16)",
        "NTLMSSP_REQUEST_TARGET (4)",
        "NTLMSSP_NEGOTIATE_OEM (2)",
        "NTLMSSP_NEGOTIATE_UNICODE (1)",
    }

    assert actual["Data"]["DomainNameFields"] == {
        "Len": 0,
        "MaxLen": 0,
        "BufferOffset": 0,
    }
    assert actual["Data"]["WorkstationFields"] == {
        "Len": 0,
        "MaxLen": 0,
        "BufferOffset": 0,
    }
    assert actual["Data"]["Version"] == {
        "Major": 10,
        "Minor": 0,
        "Build": 17763,
        "Reserved": "000000",
        "NTLMRevision": 15,
    }
    assert actual["Data"]["Payload"] == {
        "DomainName": None,
        "Workstation": None,
    }


def test_ntlm_challenge(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("ntlm_challenge")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "CHALLENGE_MESSAGE (2)"
    assert actual["RawData"] is not None

    assert actual["Data"]["TargetNameFields"] == {
        "Len": 12,
        "MaxLen": 12,
        "BufferOffset": 56,
    }
    assert actual["Data"]["NegotiateFlags"]["raw"] == 3800662581
    assert set(actual["Data"]["NegotiateFlags"]["flags"]) == {
        "NTLMSSP_NEGOTIATE_56 (2147483648)",
        "NTLMSSP_NEGOTIATE_KEY_EXCH (1073741824)",
        "NTLMSSP_NEGOTIATE_128 (536870912)",
        "NTLMSSP_NEGOTIATE_VERSION (33554432)",
        "NTLMSSP_NEGOTIATE_TARGET_INFO (8388608)",
        "NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY (524288)",
        "NTLMSSP_TARGET_TYPE_DOMAIN (65536)",
        "NTLMSSP_NEGOTIATE_ALWAYS_SIGN (32768)",
        "NTLMSSP_NEGOTIATE_NTLM (512)",
        "NTLMSSP_NEGOTIATE_SEAL (32)",
        "NTLMSSP_NEGOTIATE_SIGN (16)",
        "NTLMSSP_REQUEST_TARGET (4)",
        "NTLMSSP_NEGOTIATE_UNICODE (1)",
    }
    assert actual["Data"]["ServerChallenge"] == "E4101014CF8A90BE"
    assert actual["Data"]["Reserved"] == "0000000000000000"
    assert actual["Data"]["TargetInfoFields"] == {
        "Len": 138,
        "MaxLen": 138,
        "BufferOffset": 68,
    }
    assert actual["Data"]["Version"] == {
        "Major": 10,
        "Minor": 0,
        "Build": 14393,
        "Reserved": "000000",
        "NTLMRevision": 15,
    }
    assert actual["Data"]["Payload"]["TargetName"] == "DOMAIN"
    assert actual["Data"]["Payload"]["TargetInfo"] == [
        {"AvId": "MSV_AV_NB_DOMAIN_NAME (2)", "Value": "DOMAIN"},
        {"AvId": "MSV_AV_NB_COMPUTER_NAME (1)", "Value": "DC01"},
        {"AvId": "MSV_AV_DNS_DOMAIN_NAME (4)", "Value": "domain.local"},
        {"AvId": "MSV_AV_DNS_COMPUTER_NAME (3)", "Value": "DC01.domain.local"},
        {"AvId": "MSV_AV_DNS_TREE_NAME (5)", "Value": "domain.local"},
        {"AvId": "MSV_AV_TIMESTAMP (7)", "Value": "2020-04-30T02:46:22.4140792Z"},
        {"AvId": "MSV_AV_EOL (0)", "Value": None},
    ]


def test_ntlm_authenticate(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("ntlm_authenticate")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "AUTHENTICATE_MESSAGE (3)"
    assert actual["RawData"] is not None

    assert actual["Data"]["LmChallengeResponseFields"] == {
        "Len": 0,
        "MaxLen": 0,
        "BufferOffset": 72,
    }
    assert actual["Data"]["NtChallengeResponseFields"] == {
        "Len": 252,
        "MaxLen": 252,
        "BufferOffset": 72,
    }
    assert actual["Data"]["DomainNameFields"] == {
        "Len": 0,
        "MaxLen": 0,
        "BufferOffset": 0,
    }
    assert actual["Data"]["UserNameFields"] == {
        "Len": 54,
        "MaxLen": 54,
        "BufferOffset": 324,
    }
    assert actual["Data"]["WorkstationFields"] == {
        "Len": 26,
        "MaxLen": 26,
        "BufferOffset": 378,
    }
    assert actual["Data"]["EncryptedRandomSessionKeyFields"] == {
        "Len": 16,
        "MaxLen": 16,
        "BufferOffset": 404,
    }
    assert actual["Data"]["NegotiateFlags"]["raw"] == 3800662581
    assert set(actual["Data"]["NegotiateFlags"]["flags"]) == {
        "NTLMSSP_NEGOTIATE_56 (2147483648)",
        "NTLMSSP_NEGOTIATE_KEY_EXCH (1073741824)",
        "NTLMSSP_NEGOTIATE_128 (536870912)",
        "NTLMSSP_NEGOTIATE_VERSION (33554432)",
        "NTLMSSP_NEGOTIATE_TARGET_INFO (8388608)",
        "NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY (524288)",
        "NTLMSSP_TARGET_TYPE_DOMAIN (65536)",
        "NTLMSSP_NEGOTIATE_ALWAYS_SIGN (32768)",
        "NTLMSSP_NEGOTIATE_NTLM (512)",
        "NTLMSSP_NEGOTIATE_SEAL (32)",
        "NTLMSSP_NEGOTIATE_SIGN (16)",
        "NTLMSSP_REQUEST_TARGET (4)",
        "NTLMSSP_NEGOTIATE_UNICODE (1)",
    }
    assert actual["Data"]["Version"] == {
        "Major": 6,
        "Minor": 2,
        "Build": 0,
        "Reserved": "000000",
        "NTLMRevision": 15,
    }
    assert actual["Data"]["MIC"] is None
    assert actual["Data"]["Payload"]["LmChallengeResponse"] is None
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["ResponseType"] == "NTLMv2"
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["NTProofStr"] == "C543CC0D1A0FBD9EC05E1AAB0771B124"
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["ClientChallenge"] == {
        "RespType": 1,
        "HiRespType": 1,
        "Reserved1": 0,
        "Reserved2": 0,
        "TimeStamp": "2020-04-30T02:46:22.4140792Z",
        "ChallengeFromClient": "DC5A7473AC5672FC",
        "Reserved3": 0,
        "AvPairs": [
            {"AvId": "MSV_AV_NB_COMPUTER_NAME (1)", "Value": "DC01"},
            {"AvId": "MSV_AV_NB_DOMAIN_NAME (2)", "Value": "DOMAIN"},
            {"AvId": "MSV_AV_DNS_COMPUTER_NAME (3)", "Value": "DC01.domain.local"},
            {"AvId": "MSV_AV_DNS_DOMAIN_NAME (4)", "Value": "domain.local"},
            {"AvId": "MSV_AV_DNS_TREE_NAME (5)", "Value": "domain.local"},
            {"AvId": "MSV_AV_FLAGS (6)", "Value": {"raw": 0, "flags": []}},
            {"AvId": "MSV_AV_TIMESTAMP (7)", "Value": "2020-04-30T02:46:22.4140792Z"},
            {"AvId": "MSV_AV_TARGET_NAME (9)", "Value": "dc01.domain.local"},
            {"AvId": "MSV_AV_CHANNEL_BINDINGS (10)", "Value": "00000000000000000000000000000000"},
            {"AvId": "MSV_AV_EOL (0)", "Value": None},
        ],
        "Reserved4": 0,
    }
    assert actual["Data"]["Payload"]["DomainName"] is None
    assert actual["Data"]["Payload"]["UserName"] == "vagrant-domain@DOMAIN.LOCAL"
    assert actual["Data"]["Payload"]["Workstation"] == "JBOREAN-LINUX"
    assert actual["Data"]["Payload"]["EncryptedRandomSessionKey"] == "3765D5AFD13EC3A6C97C0726CE9F30C9"
    assert actual["Data"]["SessionKey"] == "Failed to derive"


def test_ntlm_authenticate_with_secret(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("ntlm_authenticate")).decode(),
            "--secret",
            "VagrantPass1",
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)
    assert actual["Data"]["SessionKey"] == "4E10F9B8AA8C2A5DF0A2AD75A8ECF433"


def test_ntlm_authenticate_ntlmv1(capsys):
    data = base64.b64encode(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x6C\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x84\x00\x00\x00"
        b"\x0C\x00"
        b"\x0C\x00"
        b"\x48\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x54\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x5C\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x9C\x00\x00\x00"
        b"\x35\x82\x80\xE2"
        b"\x05\x01\x28\x0A\x00\x00\x00\x0F"
        b"\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x55\x00\x73\x00\x65\x00\x72\x00"
        b"\x43\x00\x4F\x00\x4D\x00\x50\x00\x55\x00\x54\x00\x45\x00\x52\x00"
        b"\x98\xDE\xF7\xB8\x7F\x88\xAA\x5D\xAF\xE2\xDF\x77\x96\x88\xA1\x72"
        b"\xDE\xF1\x1C\x7D\x5C\xCD\xEF\x13"
        b"\x67\xC4\x30\x11\xF3\x02\x98\xA2\xAD\x35\xEC\xE6\x4F\x16\x33\x1C"
        b"\x44\xBD\xBE\xD9\x27\x84\x1F\x94"
        b"\x51\x88\x22\xB1\xB3\xF3\x50\xC8\x95\x86\x82\xEC\xBB\x3E\x3C\xB7"
    ).decode()

    entrypoint.main(["--token", data])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["Data"]["LmChallengeResponseFields"] == {
        "Len": 24,
        "MaxLen": 24,
        "BufferOffset": 108,
    }
    assert actual["Data"]["NtChallengeResponseFields"] == {
        "Len": 24,
        "MaxLen": 24,
        "BufferOffset": 132,
    }

    assert actual["Data"]["Payload"]["LmChallengeResponse"]["ResponseType"] == "LMv1"
    assert (
        actual["Data"]["Payload"]["LmChallengeResponse"]["LMProofStr"]
        == "98DEF7B87F88AA5DAFE2DF779688A172DEF11C7D5CCDEF13"
    )
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["ResponseType"] == "NTLMv1"
    assert (
        actual["Data"]["Payload"]["NtChallengeResponse"]["NTProofStr"]
        == "67C43011F30298A2AD35ECE64F16331C44BDBED927841F94"
    )


def test_ntlm_authenticate_lmv2(capsys):
    data = base64.b64encode(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x6C\x00\x00\x00"
        b"\x54\x00"
        b"\x54\x00"
        b"\x84\x00\x00\x00"
        b"\x0C\x00"
        b"\x0C\x00"
        b"\x48\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x54\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x5C\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\xD8\x00\x00\x00"
        b"\x35\x82\x88\xE2"
        b"\x05\x01\x28\x0A\x00\x00\x00\x0F"
        b"\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x55\x00\x73\x00\x65\x00\x72\x00"
        b"\x43\x00\x4F\x00\x4D\x00\x50\x00\x55\x00\x54\x00\x45\x00\x52\x00"
        b"\x86\xC3\x50\x97\xAC\x9C\xEC\x10\x25\x54\x76\x4A\x57\xCC\xCC\x19"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x68\xCD\x0A\xB8\x51\xE5\x1C\x96\xAA\xBC\x92\x7B\xEB\xEF\x6A\x1C"
        b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x00\x00\x00\x00"
        b"\x02\x00\x0C\x00\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x01\x00\x0C\x00\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\xC5\xDA\xD2\x54\x4F\xC9\x79\x90\x94\xCE\x1C\xE9\x0B\xC9\xD0\x3E"
    ).decode()

    entrypoint.main(["--token", data])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["Data"]["LmChallengeResponseFields"] == {
        "Len": 24,
        "MaxLen": 24,
        "BufferOffset": 108,
    }
    assert actual["Data"]["NtChallengeResponseFields"] == {
        "Len": 84,
        "MaxLen": 84,
        "BufferOffset": 132,
    }

    assert actual["Data"]["Payload"]["LmChallengeResponse"]["ResponseType"] == "LMv2"
    assert actual["Data"]["Payload"]["LmChallengeResponse"]["LMProofStr"] == "86C35097AC9CEC102554764A57CCCC19"
    assert actual["Data"]["Payload"]["LmChallengeResponse"]["ChallengeFromClient"] == "AAAAAAAAAAAAAAAA"
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["ResponseType"] == "NTLMv2"
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["NTProofStr"] == "68CD0AB851E51C96AABC927BEBEF6A1C"
    assert actual["Data"]["Payload"]["NtChallengeResponse"]["ClientChallenge"] == {
        "RespType": 1,
        "HiRespType": 1,
        "Reserved1": 0,
        "Reserved2": 0,
        "TimeStamp": "1601-01-01T00:00:00Z",
        "ChallengeFromClient": "AAAAAAAAAAAAAAAA",
        "Reserved3": 0,
        "AvPairs": [
            {"AvId": "MSV_AV_NB_DOMAIN_NAME (2)", "Value": "Domain"},
            {"AvId": "MSV_AV_NB_COMPUTER_NAME (1)", "Value": "Server"},
            {"AvId": "MSV_AV_EOL (0)", "Value": None},
        ],
        "Reserved4": 0,
    }


def test_nltm_session_key_no_sign_seal(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("ntlm_authenticate_no_sign_seal")).decode(),
            "--secret",
            "vagrant",
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)
    assert actual["Data"]["SessionKey"] == "92D93576421343CEB1937AAD4CD78116"


def test_ntlm_without_version(capsys):
    msg = Negotiate()
    entrypoint.main(
        [
            "--token",
            base64.b64encode(msg.pack()).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["Data"]["Version"] is None


def test_ntlm_no_target_info(capsys):
    msg = Challenge()
    entrypoint.main(
        [
            "--token",
            base64.b64encode(msg.pack()).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["Data"]["Payload"]["TargetInfo"] is None


def test_ntlm_av_single_host(capsys):
    ti = TargetInfo(
        {
            AvId.single_host: SingleHost(size=4, z4=0, custom_data=b"\x01" * 8, machine_id=b"\x02" * 32),
        }
    )
    msg = Challenge(target_info=ti)
    entrypoint.main(
        [
            "--token",
            base64.b64encode(msg.pack()).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert len(actual["Data"]["Payload"]["TargetInfo"]) == 2
    assert actual["Data"]["Payload"]["TargetInfo"][0]["AvId"] == "MSV_AV_SINGLE_HOST (8)"
    assert actual["Data"]["Payload"]["TargetInfo"][0]["Value"] == {
        "Size": 4,
        "Z4": 0,
        "CustomData": "0101010101010101",
        "MachineId": "0202020202020202020202020202020202020202020202020202020202020202",
    }
    assert actual["Data"]["Payload"]["TargetInfo"][1] == {"AvId": "MSV_AV_EOL (0)", "Value": None}


def test_krb_as_rep(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("krb_as_rep")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "AS-REP (11)"
    assert actual["RawData"] is not None

    assert actual["Data"]["pvno"] == 5
    assert actual["Data"]["msg-type"] == "AS-REP (11)"
    assert actual["Data"]["padata"] == [
        {
            "padata-type": "PA-ETYPE-INFO2 (19)",
            "padata-value": [
                {
                    "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
                    "salt": "444F4D41494E2E4C4F43414C76616772616E742D646F6D61696E",
                    "s2kparams": None,
                },
            ],
        },
    ]
    assert actual["Data"]["crealm"] == "DOMAIN.LOCAL"
    assert actual["Data"]["cname"] == {
        "name-type": "NT-PRINCIPAL (1)",
        "name-string": ["vagrant-domain"],
    }
    assert actual["Data"]["ticket"] == {
        "tkt-vno": 5,
        "realm": "DOMAIN.LOCAL",
        "sname": {
            "name-type": "NT-SRV-INST (2)",
            "name-string": ["krbtgt", "DOMAIN.LOCAL"],
        },
        "enc-part": {
            "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
            "kvno": 2,
            "cipher": actual["Data"]["ticket"]["enc-part"]["cipher"],
        },
    }
    assert actual["Data"]["ticket"]["enc-part"]["cipher"] is not None
    assert actual["Data"]["enc-part"] == {
        "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
        "kvno": 11,
        "cipher": actual["Data"]["enc-part"]["cipher"],
    }
    assert actual["Data"]["enc-part"]["cipher"] is not None


def test_krb_as_req(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("krb_as_req")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "AS-REQ (10)"
    assert actual["RawData"] is not None

    assert actual["Data"]["pvno"] == 5
    assert actual["Data"]["msg-type"] == "AS-REQ (10)"
    assert actual["Data"]["padata"] == [
        {
            "padata-type": "PA-ENC-TIMESTAMP (2)",
            "padata-value": {
                "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
                "kvno": None,
                "cipher": actual["Data"]["padata"][0]["padata-value"]["cipher"],
            },
        },
        {
            "padata-type": "PA-REQ-ENC-PA-REP (149)",
            "padata-value": "",
        },
    ]
    assert actual["Data"]["padata"][0]["padata-value"]["cipher"] is not None
    assert actual["Data"]["req-body"] == {
        "kdc-options": {
            "raw": 1073741824,
            "flags": ["forwardable (1073741824)"],
        },
        "cname": {
            "name-type": "NT-PRINCIPAL (1)",
            "name-string": ["vagrant-domain"],
        },
        "realm": "DOMAIN.LOCAL",
        "sname": {
            "name-type": "NT-SRV-INST (2)",
            "name-string": ["krbtgt", "DOMAIN.LOCAL"],
        },
        "from": None,
        "till": "2020-06-14T07:04:20+00:00",
        "rtime": None,
        "nonce": 734266074,
        "etype": [
            "AES256_CTS_HMAC_SHA1_96 (18)",
            "AES128_CTS_HMAC_SHA1_96 (17)",
            "DES3_CBC_SHA1 (16)",
            "RC4_HMAC (23)",
        ],
        "addresses": None,
        "enc-authorization-data": None,
        "additional-tickets": None,
    }


def test_krb_error(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("krb_error")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "KRB-ERROR (30)"
    assert actual["RawData"] is not None

    assert actual["Data"]["pvno"] == 5
    assert actual["Data"]["msg-type"] == "KRB-ERROR (30)"
    assert actual["Data"]["ctime"] is None
    assert actual["Data"]["cusec"] is None
    assert actual["Data"]["stime"] == "2020-06-13T21:04:23+00:00"
    assert actual["Data"]["susec"] == 748591
    assert actual["Data"]["error-code"] == "KDC_ERR_PREAUTH_REQUIRED (25)"
    assert actual["Data"]["crealm"] is None
    assert actual["Data"]["cname"] is None
    assert actual["Data"]["realm"] == "DOMAIN.LOCAL"
    assert actual["Data"]["sname"] == {"name-type": "NT-SRV-INST (2)", "name-string": ["krbtgt", "DOMAIN.LOCAL"]}
    assert actual["Data"]["e-text"] is None
    assert actual["Data"]["e-data"] is not None


def test_krb_tgs_rep(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("krb_tgs_rep")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "TGS-REP (13)"
    assert actual["RawData"] is not None

    assert actual["Data"]["pvno"] == 5
    assert actual["Data"]["msg-type"] == "TGS-REP (13)"
    assert actual["Data"]["padata"] is None
    assert actual["Data"]["crealm"] == "DOMAIN.LOCAL"
    assert actual["Data"]["cname"] == {
        "name-type": "NT-PRINCIPAL (1)",
        "name-string": ["vagrant-domain"],
    }
    assert actual["Data"]["ticket"] == {
        "tkt-vno": 5,
        "realm": "DOMAIN.LOCAL",
        "sname": {
            "name-type": "NT-SRV-HST (3)",
            "name-string": ["HTTP", "server2019.domain.local"],
        },
        "enc-part": {
            "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
            "kvno": 6,
            "cipher": actual["Data"]["ticket"]["enc-part"]["cipher"],
        },
    }
    assert actual["Data"]["ticket"]["enc-part"]["cipher"] is not None
    assert actual["Data"]["enc-part"] == {
        "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
        "kvno": None,
        "cipher": actual["Data"]["enc-part"]["cipher"],
    }
    assert actual["Data"]["enc-part"]["cipher"] is not None


def test_krb_tgs_req(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("krb_tgs_req")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "TGS-REQ (12)"
    assert actual["RawData"] is not None

    assert actual["Data"]["pvno"] == 5
    assert actual["Data"]["msg-type"] == "TGS-REQ (12)"
    assert len(actual["Data"]["padata"]) == 1
    assert actual["Data"]["padata"][0]["padata-type"] == "PA-TGS-REQ (1)"
    assert actual["Data"]["padata"][0]["padata-value"]["pvno"] == 5
    assert actual["Data"]["padata"][0]["padata-value"]["msg-type"] == "AP-REQ (14)"
    assert actual["Data"]["padata"][0]["padata-value"]["ap-options"] == {
        "raw": 0,
        "flags": [],
    }
    assert actual["Data"]["padata"][0]["padata-value"]["ticket"] == {
        "tkt-vno": 5,
        "realm": "DOMAIN.LOCAL",
        "sname": {"name-type": "NT-SRV-INST (2)", "name-string": ["krbtgt", "DOMAIN.LOCAL"]},
        "enc-part": {
            "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
            "kvno": 2,
            "cipher": actual["Data"]["padata"][0]["padata-value"]["ticket"]["enc-part"]["cipher"],
        },
    }
    assert actual["Data"]["padata"][0]["padata-value"]["ticket"]["enc-part"]["cipher"] is not None
    assert actual["Data"]["padata"][0]["padata-value"]["authenticator"] == {
        "etype": "AES256_CTS_HMAC_SHA1_96 (18)",
        "kvno": None,
        "cipher": actual["Data"]["padata"][0]["padata-value"]["authenticator"]["cipher"],
    }
    assert actual["Data"]["padata"][0]["padata-value"]["authenticator"]["cipher"] is not None

    assert actual["Data"]["req-body"]["kdc-options"] == {
        "raw": 1073807360,
        "flags": ["forwardable (1073741824)", "canonicalize (65536)"],
    }
    assert actual["Data"]["req-body"]["cname"] is None
    assert actual["Data"]["req-body"]["realm"] == "DOMAIN.LOCAL"
    assert actual["Data"]["req-body"]["sname"] == {
        "name-type": "NT-SRV-HST (3)",
        "name-string": ["HTTP", "server2019.domain.local"],
    }
    assert actual["Data"]["req-body"]["from"] is None
    assert actual["Data"]["req-body"]["till"] == "1970-01-01T00:00:00+00:00"
    assert actual["Data"]["req-body"]["rtime"] is None
    assert actual["Data"]["req-body"]["nonce"] == 333512069
    assert actual["Data"]["req-body"]["etype"] == [
        "AES256_CTS_HMAC_SHA1_96 (18)",
        "AES128_CTS_HMAC_SHA1_96 (17)",
        "DES3_CBC_SHA1 (16)",
        "RC4_HMAC (23)",
    ]
    assert actual["Data"]["req-body"]["addresses"] is None
    assert actual["Data"]["req-body"]["enc-authorization-data"] is None
    assert actual["Data"]["req-body"]["additional-tickets"] is None


def test_krb_ap_req_in_initial_context_token(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("initial_context_token_krb_ap_rep")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "SPNEGO InitialContextToken"
    assert actual["RawData"] is not None

    assert actual["Data"]["thisMech"] == "Kerberos (1.2.840.113554.1.2.2)"
    assert actual["Data"]["innerContextToken"]["MessageType"] == "AP-REP (15)"
    assert actual["Data"]["innerContextToken"]["RawData"] is not None
    assert actual["Data"]["innerContextToken"]["Data"]["pvno"] == 5
    assert actual["Data"]["innerContextToken"]["Data"]["msg-type"] == "AP-REP (15)"


def test_neg_token_init(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("initial_context_token_neg_token_init")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "SPNEGO InitialContextToken"
    assert actual["RawData"] is not None

    assert actual["Data"]["thisMech"] == "SPNEGO (1.3.6.1.5.5.2)"
    assert actual["Data"]["innerContextToken"]["MessageType"] == "SPNEGO NegTokenInit"
    assert actual["Data"]["innerContextToken"]["RawData"] is not None
    assert actual["Data"]["innerContextToken"]["Data"]["mechTypes"] == [
        "Kerberos (1.2.840.113554.1.2.2)",
        "NTLM (1.3.6.1.4.1.311.2.2.10)",
    ]
    assert actual["Data"]["innerContextToken"]["Data"]["reqFlags"] is None
    assert actual["Data"]["innerContextToken"]["Data"]["mechListMIC"] is None

    mech_token = actual["Data"]["innerContextToken"]["Data"]["mechToken"]
    assert mech_token is not None

    assert mech_token["MessageType"] == "SPNEGO InitialContextToken"
    assert mech_token["RawData"] is not None

    assert mech_token["Data"]["thisMech"] == "Kerberos (1.2.840.113554.1.2.2)"
    assert mech_token["Data"]["innerContextToken"]["MessageType"] == "AP-REQ (14)"


def test_neg_token_init2(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("initial_context_token_neg_token_init2")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "SPNEGO InitialContextToken"
    assert actual["RawData"] is not None

    assert actual["Data"]["thisMech"] == "SPNEGO (1.3.6.1.5.5.2)"
    assert actual["Data"]["innerContextToken"]["MessageType"] == "SPNEGO NegTokenInit2"
    assert actual["Data"]["innerContextToken"]["RawData"] is not None
    assert actual["Data"]["innerContextToken"]["Data"]["mechTypes"] == [
        "NEGOEX (1.3.6.1.4.1.311.2.2.30)",
        "MS Kerberos (1.2.840.48018.1.2.2)",
        "Kerberos (1.2.840.113554.1.2.2)",
        "Kerberos User to User (1.2.840.113554.1.2.2.3)",
        "NTLM (1.3.6.1.4.1.311.2.2.10)",
    ]
    assert actual["Data"]["innerContextToken"]["Data"]["reqFlags"] is None
    assert actual["Data"]["innerContextToken"]["Data"]["mechToken"] is None
    assert actual["Data"]["innerContextToken"]["Data"]["mechListMIC"] is None
    assert actual["Data"]["innerContextToken"]["Data"]["negHints"] == {
        "hintName": "not_defined_in_RFC4178@please_ignore",
        "hintAddress": None,
    }


def test_neg_token_resp(capsys):
    entrypoint.main(
        [
            "--token",
            base64.b64encode(get_data("neg_token_resp")).decode(),
        ]
    )
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert actual["MessageType"] == "SPNEGO NegTokenResp"
    assert actual["RawData"] is not None
    assert actual["Data"]["negState"] == "accept-complete (0)"
    assert actual["Data"]["supportedMech"] == "Kerberos (1.2.840.113554.1.2.2)"
    assert actual["Data"]["responseToken"]["MessageType"] == "SPNEGO InitialContextToken"
    assert actual["Data"]["responseToken"]["Data"]["thisMech"] == "Kerberos (1.2.840.113554.1.2.2)"
    assert actual["Data"]["responseToken"]["Data"]["innerContextToken"]["MessageType"] == "AP-REP (15)"
    assert actual["Data"]["mechListMIC"] is None


def test_tls13_client_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.3_client_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual[0]["RawData"] is not None

    actual = actual[0]["Data"]
    assert isinstance(actual, list)
    assert len(actual) == 1
    assert isinstance(actual[0], dict)

    assert actual[0]["HandshakeType"] == "client_hello (1)"
    assert actual[0]["RawData"] is not None
    actual = actual[0]["Data"]

    assert actual["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual["Random"] == "6A8DEBD01B3A24DDEE8428FDB204F819C5CC5FC9B780B037B723446D3F4CE66B"
    assert actual["SessionID"] == "2AE52188EF97D27FD6EA670EC3DB3064DBE431B12E262A0B87F45EAA13E29F45"
    assert len(actual["CipherSuites"]) == 36
    for cs in actual["CipherSuites"]:
        assert isinstance(cs, str)
    assert len(actual["CompressionMethods"]) == 1
    assert actual["CompressionMethods"][0] == "none (0)"
    assert len(actual["Extensions"]) == 10

    assert actual["Extensions"][0]["ExtensionType"] == "server_name (0)"
    assert actual["Extensions"][0]["Data"] == [{"Type": "server_name (0)", "Name": "server2022.domain.test"}]
    assert actual["Extensions"][0]["RawData"] is not None

    assert actual["Extensions"][1]["ExtensionType"] == "ec_point_formats (11)"
    assert actual["Extensions"][1]["Data"] == [
        "uncompressed (0)",
        "ansiX962_compressed_prime (1)",
        "ansiX962_compressed_char2 (2)",
    ]
    assert actual["Extensions"][1]["RawData"] is not None

    assert actual["Extensions"][2]["ExtensionType"] == "supported_groups (10)"
    assert actual["Extensions"][2]["Data"] == [
        "x25519 (29)",
        "secp256r1 (23)",
        "x448 (30)",
        "secp521r1 (25)",
        "secp384r1 (24)",
        "ffdhe2048 (256)",
        "ffdhe3072 (257)",
        "ffdhe4096 (258)",
        "ffdhe6144 (259)",
        "ffdhe8192 (260)",
    ]
    assert actual["Extensions"][2]["RawData"] is not None

    assert actual["Extensions"][3] == {"ExtensionType": "session_ticket (35)"}

    assert actual["Extensions"][4] == {"ExtensionType": "encrypt_then_mac (22)"}

    assert actual["Extensions"][5] == {"ExtensionType": "extended_master_secret (23)"}

    assert actual["Extensions"][6]["ExtensionType"] == "signature_algorithms (13)"
    assert actual["Extensions"][6]["Data"] == [
        "ecdsa_secp256r1_sha256 (1027)",
        "ecdsa_secp384r1_sha384 (1283)",
        "ecdsa_secp521r1_sha512 (1539)",
        "ed25519 (2055)",
        "ed448 (2056)",
        "rsa_pss_pss_sha256 (2057)",
        "rsa_pss_pss_sha384 (2058)",
        "rsa_pss_pss_sha512 (2059)",
        "rsa_pss_rsae_sha256 (2052)",
        "rsa_pss_rsae_sha384 (2053)",
        "rsa_pss_rsae_sha512 (2054)",
        "rsa_pkcs1_sha256 (1025)",
        "rsa_pkcs1_sha384 (1281)",
        "rsa_pkcs1_sha512 (1537)",
        "sha224_ecdsa (771)",
        "sha224_rsa (769)",
    ]
    assert actual["Extensions"][6]["RawData"] is not None

    assert actual["Extensions"][7]["ExtensionType"] == "supported_versions (43)"
    assert actual["Extensions"][7]["Data"] == [
        "TLS 1.3 (0x0304) (772)",
        "TLS 1.2 (0x0303) (771)",
    ]
    assert actual["Extensions"][7]["RawData"] is not None

    assert actual["Extensions"][8]["ExtensionType"] == "psk_key_exchange_modes (45)"
    assert actual["Extensions"][8]["Data"] == [
        "psk_dhe_ke (1)",
    ]
    assert actual["Extensions"][8]["RawData"] is not None

    assert actual["Extensions"][9]["ExtensionType"] == "key_share (51)"
    assert actual["Extensions"][9]["Data"] == [
        {
            "Group": "x25519 (29)",
            "Key": "C016598ACACC7390AC94D0479A20B490D8FDDA0D7CFBA4F86A8996098B96293F",
        },
    ]
    assert actual["Extensions"][9]["RawData"] is not None


def test_tls12_client_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.2_client_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual[0]["RawData"] is not None

    actual = actual[0]["Data"]
    assert isinstance(actual, list)
    assert len(actual) == 1
    assert isinstance(actual[0], dict)

    assert actual[0]["HandshakeType"] == "client_hello (1)"
    assert actual[0]["RawData"] is not None
    actual = actual[0]["Data"]

    assert actual["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual["Random"] == "003E759A1FC0B769CEF34EE66CA42CE37AD626D2B7D27CACB590358D642EE878"
    assert actual["SessionID"] == ""
    assert len(actual["CipherSuites"]) == 32
    for cs in actual["CipherSuites"]:
        assert isinstance(cs, str)
    assert len(actual["CompressionMethods"]) == 1
    assert actual["CompressionMethods"][0] == "none (0)"
    assert len(actual["Extensions"]) == 7

    assert actual["Extensions"][0]["ExtensionType"] == "server_name (0)"
    assert actual["Extensions"][0]["Data"] == [{"Type": "server_name (0)", "Name": "server2022.domain.test"}]
    assert actual["Extensions"][0]["RawData"] is not None

    assert actual["Extensions"][1]["ExtensionType"] == "ec_point_formats (11)"
    assert actual["Extensions"][1]["Data"] == [
        "uncompressed (0)",
        "ansiX962_compressed_prime (1)",
        "ansiX962_compressed_char2 (2)",
    ]
    assert actual["Extensions"][1]["RawData"] is not None

    assert actual["Extensions"][2]["ExtensionType"] == "supported_groups (10)"
    assert actual["Extensions"][2]["Data"] == [
        "x25519 (29)",
        "secp256r1 (23)",
        "x448 (30)",
        "secp521r1 (25)",
        "secp384r1 (24)",
    ]
    assert actual["Extensions"][2]["RawData"] is not None

    assert actual["Extensions"][3] == {"ExtensionType": "session_ticket (35)"}

    assert actual["Extensions"][4] == {"ExtensionType": "encrypt_then_mac (22)"}

    assert actual["Extensions"][5] == {"ExtensionType": "extended_master_secret (23)"}

    assert actual["Extensions"][6]["ExtensionType"] == "signature_algorithms (13)"
    assert actual["Extensions"][6]["Data"] == [
        "ecdsa_secp256r1_sha256 (1027)",
        "ecdsa_secp384r1_sha384 (1283)",
        "ecdsa_secp521r1_sha512 (1539)",
        "ed25519 (2055)",
        "ed448 (2056)",
        "rsa_pss_pss_sha256 (2057)",
        "rsa_pss_pss_sha384 (2058)",
        "rsa_pss_pss_sha512 (2059)",
        "rsa_pss_rsae_sha256 (2052)",
        "rsa_pss_rsae_sha384 (2053)",
        "rsa_pss_rsae_sha512 (2054)",
        "rsa_pkcs1_sha256 (1025)",
        "rsa_pkcs1_sha384 (1281)",
        "rsa_pkcs1_sha512 (1537)",
        "sha224_ecdsa (771)",
        "sha224_rsa (769)",
    ]
    assert actual["Extensions"][6]["RawData"] is not None


def test_tls11_client_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.1_client_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual[0]["RawData"] is not None

    actual = actual[0]["Data"]
    assert isinstance(actual, list)
    assert len(actual) == 1
    assert isinstance(actual[0], dict)

    assert actual[0]["HandshakeType"] == "client_hello (1)"
    assert actual[0]["RawData"] is not None
    actual = actual[0]["Data"]

    assert actual["ProtocolVersion"] == "TLS 1.1 (0x0302) (770)"
    assert actual["Random"] == "2DBA71B73F9355CA456C2BE070BAEF289EC194678A18BA6467DA96D1211C45CC"
    assert actual["SessionID"] == ""
    assert len(actual["CipherSuites"]) == 9
    for cs in actual["CipherSuites"]:
        assert isinstance(cs, str)
    assert len(actual["CompressionMethods"]) == 1
    assert actual["CompressionMethods"][0] == "none (0)"
    assert len(actual["Extensions"]) == 6

    assert actual["Extensions"][0]["ExtensionType"] == "server_name (0)"
    assert actual["Extensions"][0]["Data"] == [{"Type": "server_name (0)", "Name": "server2022.domain.test"}]
    assert actual["Extensions"][0]["RawData"] is not None

    assert actual["Extensions"][1]["ExtensionType"] == "ec_point_formats (11)"
    assert actual["Extensions"][1]["Data"] == [
        "uncompressed (0)",
        "ansiX962_compressed_prime (1)",
        "ansiX962_compressed_char2 (2)",
    ]
    assert actual["Extensions"][1]["RawData"] is not None

    assert actual["Extensions"][2]["ExtensionType"] == "supported_groups (10)"
    assert actual["Extensions"][2]["Data"] == [
        "x25519 (29)",
        "secp256r1 (23)",
        "x448 (30)",
        "secp521r1 (25)",
        "secp384r1 (24)",
    ]
    assert actual["Extensions"][2]["RawData"] is not None

    assert actual["Extensions"][3] == {"ExtensionType": "session_ticket (35)"}

    assert actual["Extensions"][4] == {"ExtensionType": "encrypt_then_mac (22)"}

    assert actual["Extensions"][5] == {"ExtensionType": "extended_master_secret (23)"}


def test_tls10_client_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.0_client_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual[0]["RawData"] is not None

    actual = actual[0]["Data"]
    assert isinstance(actual, list)
    assert len(actual) == 1
    assert isinstance(actual[0], dict)

    assert actual[0]["HandshakeType"] == "client_hello (1)"
    assert actual[0]["RawData"] is not None
    actual = actual[0]["Data"]

    assert actual["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual["Random"] == "009B6CD8070904706A6CD7AD711D04E942FF64500853D0CCB7709285ED0D31F5"
    assert actual["SessionID"] == ""
    assert len(actual["CipherSuites"]) == 9
    for cs in actual["CipherSuites"]:
        assert isinstance(cs, str)
    assert len(actual["CompressionMethods"]) == 1
    assert actual["CompressionMethods"][0] == "none (0)"
    assert len(actual["Extensions"]) == 6

    assert actual["Extensions"][0]["ExtensionType"] == "server_name (0)"
    assert actual["Extensions"][0]["Data"] == [{"Type": "server_name (0)", "Name": "server2022.domain.test"}]
    assert actual["Extensions"][0]["RawData"] is not None

    assert actual["Extensions"][1]["ExtensionType"] == "ec_point_formats (11)"
    assert actual["Extensions"][1]["Data"] == [
        "uncompressed (0)",
        "ansiX962_compressed_prime (1)",
        "ansiX962_compressed_char2 (2)",
    ]
    assert actual["Extensions"][1]["RawData"] is not None

    assert actual["Extensions"][2]["ExtensionType"] == "supported_groups (10)"
    assert actual["Extensions"][2]["Data"] == [
        "x25519 (29)",
        "secp256r1 (23)",
        "x448 (30)",
        "secp521r1 (25)",
        "secp384r1 (24)",
    ]
    assert actual["Extensions"][2]["RawData"] is not None

    assert actual["Extensions"][3] == {"ExtensionType": "session_ticket (35)"}

    assert actual["Extensions"][4] == {"ExtensionType": "encrypt_then_mac (22)"}

    assert actual["Extensions"][5] == {"ExtensionType": "extended_master_secret (23)"}


def test_tls13_server_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.3_server_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 3

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[0]["RawData"] is not None

    actual_handshakes = actual[0]["Data"]
    assert isinstance(actual_handshakes, list)
    assert len(actual_handshakes) == 1
    assert isinstance(actual_handshakes[0], dict)

    assert actual_handshakes[0]["HandshakeType"] == "server_hello (2)"
    assert actual_handshakes[0]["RawData"] is not None

    actual_hello = actual_handshakes[0]["Data"]
    assert actual_hello["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual_hello["Random"] == "0C330D260EFB17B9BC142F29ACA1369B5176ACE30BFBAE56090DEC68A59BF457"
    assert actual_hello["SessionID"] == "2AE52188EF97D27FD6EA670EC3DB3064DBE431B12E262A0B87F45EAA13E29F45"
    assert actual_hello["CipherSuite"] == "TLS_AES_256_GCM_SHA384 - 0x1302"
    assert actual_hello["CompressionMethod"] == "none (0)"
    assert len(actual_hello["Extensions"]) == 2

    assert actual_hello["Extensions"][0]["ExtensionType"] == "supported_versions (43)"
    assert actual_hello["Extensions"][0]["Data"] == "TLS 1.3 (0x0304) (772)"
    assert actual_hello["Extensions"][0]["RawData"] is not None

    assert actual_hello["Extensions"][1]["ExtensionType"] == "key_share (51)"
    assert actual_hello["Extensions"][1]["Data"] == {
        "Group": "x25519 (29)",
        "Key": "20BABC51697B813B09E7228F3453178C35A1D4005584859EEB5DC38248C84950",
    }
    assert actual_hello["Extensions"][1]["RawData"] is not None

    assert isinstance(actual[1], dict)
    assert actual[1] == {
        "ContentType": "change_cipher_spec (20)",
        "ProtocolVersion": "TLS 1.2 (0x0303) (771)",
        "RawData": "140303000101",
    }

    assert isinstance(actual[2], dict)
    assert actual[2]["ContentType"] == "application_data (23)"
    assert actual[2]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[2]["RawData"] is not None


def test_tls12_server_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.2_server_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 4

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[0]["RawData"] is not None

    entry1 = actual[0]["Data"]
    assert isinstance(entry1, list)
    assert len(entry1) == 1
    assert isinstance(entry1[0], dict)

    assert entry1[0]["HandshakeType"] == "server_hello (2)"
    assert entry1[0]["RawData"] is not None

    actual_hello = entry1[0]["Data"]
    assert actual_hello["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual_hello["Random"] == "3A1EAF5E3C0B9BD4FE889BAE819C1F392613D5320910907469C7A79158430C40"
    assert actual_hello["SessionID"] == "394772ADC4255C0B50C4F3892CE8AC4827A63F1FE056E1E60B35306F83B19A27"
    assert actual_hello["CipherSuite"] == "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 - 0xC02F"
    assert actual_hello["CompressionMethod"] == "none (0)"
    assert len(actual_hello["Extensions"]) == 4

    assert actual_hello["Extensions"][0] == {"ExtensionType": "server_name (0)"}

    assert actual_hello["Extensions"][1]["ExtensionType"] == "renegotiation_info (65281)"
    assert actual_hello["Extensions"][1]["RawData"] == "00"

    assert actual_hello["Extensions"][2]["ExtensionType"] == "ec_point_formats (11)"
    assert actual_hello["Extensions"][2]["Data"] == [
        "uncompressed (0)",
        "ansiX962_compressed_prime (1)",
        "ansiX962_compressed_char2 (2)",
    ]
    assert actual_hello["Extensions"][2]["RawData"] == "03000102"

    assert actual_hello["Extensions"][3]["ExtensionType"] == "application_layer_protocol_negotiation (16)"
    assert actual_hello["Extensions"][3]["Data"] == [
        "http/1.1",
    ]
    assert actual_hello["Extensions"][3]["RawData"] == "000908687474702F312E31"

    assert isinstance(actual[1], dict)
    assert actual[1]["ContentType"] == "handshake (22)"
    assert actual[1]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[1]["RawData"] is not None

    entry2 = actual[1]["Data"]
    assert isinstance(entry2, list)
    assert len(entry2) == 1

    assert isinstance(entry2[0], dict)
    assert entry2[0]["HandshakeType"] == "certificate (11)"
    assert entry2[0]["RawData"] is not None

    actual_certificate = entry2[0]["Data"]
    assert actual_certificate["Certificate"] is not None

    assert isinstance(actual[2], dict)
    assert actual[2]["ContentType"] == "handshake (22)"
    assert actual[2]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[2]["RawData"] is not None

    entry3 = actual[2]["Data"]
    assert isinstance(entry3, list)
    assert len(entry3) == 1

    assert isinstance(entry2[0], dict)
    assert entry3[0]["HandshakeType"] == "server_key_exchange (12)"
    assert entry3[0]["RawData"] is not None

    actual_server_key_exch = entry3[0]["Data"]
    assert actual_server_key_exch["CurveType"] == "named_curve (3)"
    assert actual_server_key_exch["Curve"] == "secp256r1 (23)"
    assert actual_server_key_exch["PublicKey"] is not None
    assert actual_server_key_exch["SignatureAlgorithm"] == "rsa_pkcs1_sha512 (1537)"
    assert actual_server_key_exch["Signature"] is not None

    assert isinstance(actual[3], dict)
    assert actual[3]["ContentType"] == "handshake (22)"
    assert actual[3]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[3]["RawData"] is not None

    entry4 = actual[3]["Data"]
    assert isinstance(entry4, list)
    assert len(entry4) == 2

    assert isinstance(entry4[0], dict)
    assert entry4[0]["HandshakeType"] == "certificate_request (13)"
    assert entry4[0]["RawData"] is not None

    actual_cert_req = entry4[0]["Data"]
    assert actual_cert_req["CertificateTypes"] == [
        "rsa_sign (1)",
        "dss_sign (2)",
        "ecdsa_sign (64)",
    ]
    assert actual_cert_req["SignatureAlgorithms"] == [
        "rsa_pkcs1_sha512 (1537)",
        "dsa_sha512 (1538)",
        "ecdsa_secp521r1_sha512 (1539)",
        "rsa_pkcs1_sha384 (1281)",
        "dsa_sha384 (1282)",
        "ecdsa_secp384r1_sha384 (1283)",
        "rsa_pkcs1_sha256 (1025)",
        "dsa_sha256 (1026)",
        "ecdsa_secp256r1_sha256 (1027)",
        "sha224_rsa (769)",
        "dsa_sha224 (770)",
        "sha224_ecdsa (771)",
        "rsa_pkcs1_sha1 (513)",
        "dsa_sha1 (514)",
        "ecdsa_sha1 (515)",
    ]
    assert actual_cert_req["CertificateAuthorities"] == [
        {"OID": "id-at-countryName (2.5.4.6)", "Value": "US"},
        {"OID": "id-at-stateOrProvinceName (2.5.4.8)", "Value": "California"},
        {"OID": "id-at-localityName (2.5.4.7)", "Value": "San Francisco"},
        {"OID": "id-at-organizationName (2.5.4.10)", "Value": "BadSSL"},
        {"OID": "id-at-commonName (2.5.4.3)", "Value": "BadSSL Client Root Certificate Authority"},
    ]

    assert isinstance(entry4[1], dict)
    assert entry4[1]["HandshakeType"] == "server_hello_done (14)"
    assert entry4[1]["RawData"] is not None


def test_tls11_server_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.1_server_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.1 (0x0302) (770)"
    assert actual[0]["RawData"] is not None

    actual_handshakes = actual[0]["Data"]
    assert isinstance(actual_handshakes, list)
    assert len(actual_handshakes) == 4
    assert isinstance(actual_handshakes[0], dict)

    assert actual_handshakes[0]["HandshakeType"] == "server_hello (2)"
    assert actual_handshakes[0]["RawData"] is not None

    actual_hello = actual_handshakes[0]["Data"]
    assert actual_hello["ProtocolVersion"] == "TLS 1.1 (0x0302) (770)"
    assert actual_hello["Random"] == "636B6A884D89DB703FB99037847AF10F7BC64846CE4EBA7C444F574E47524400"
    assert actual_hello["SessionID"] == "2610000040B449AC2B16836E336AF8E42783E74EE2E2ABB2D5916199C996B633"
    assert actual_hello["CipherSuite"] == "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA - 0xC014"
    assert actual_hello["CompressionMethod"] == "none (0)"
    assert len(actual_hello["Extensions"]) == 2

    assert actual_hello["Extensions"][0] == {
        "ExtensionType": "extended_master_secret (23)",
    }

    assert actual_hello["Extensions"][1]["ExtensionType"] == "renegotiation_info (65281)"
    assert actual_hello["Extensions"][1]["RawData"] == "00"

    assert actual_handshakes[1]["HandshakeType"] == "certificate (11)"
    assert actual_handshakes[1]["RawData"] is not None
    actual_certificate = actual_handshakes[1]["Data"]
    assert actual_certificate["Certificate"] is not None

    assert actual_handshakes[2]["HandshakeType"] == "server_key_exchange (12)"
    assert actual_handshakes[2]["RawData"] is not None
    actual_server_key_exch = actual_handshakes[2]["Data"]
    assert actual_server_key_exch["CurveType"] == "named_curve (3)"
    assert actual_server_key_exch["Curve"] == "secp384r1 (24)"
    assert actual_server_key_exch["PublicKey"] is not None
    assert actual_server_key_exch["SignatureAlgorithm"] is None
    assert actual_server_key_exch["Signature"] is not None

    assert actual_handshakes[3]["HandshakeType"] == "server_hello_done (14)"
    assert actual_handshakes[3]["RawData"] is not None


def test_tls10_server_hello(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.0_server_hello")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual[0]["RawData"] is not None

    actual_handshakes = actual[0]["Data"]
    assert isinstance(actual_handshakes, list)
    assert len(actual_handshakes) == 4
    assert isinstance(actual_handshakes[0], dict)

    assert actual_handshakes[0]["HandshakeType"] == "server_hello (2)"
    assert actual_handshakes[0]["RawData"] is not None

    actual_hello = actual_handshakes[0]["Data"]
    assert actual_hello["ProtocolVersion"] == "TLS 1.0 (0x0301) (769)"
    assert actual_hello["Random"] == "636B6A8EF194ACC4C15811DBEC314A07D8ECF15DEF047781444F574E47524400"
    assert actual_hello["SessionID"] == "2B380000A583E220707B9E7D5372F3ED2946D32586F79E5BC0E36BD4DD8E6A0C"
    assert actual_hello["CipherSuite"] == "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA - 0xC014"
    assert actual_hello["CompressionMethod"] == "none (0)"
    assert len(actual_hello["Extensions"]) == 2

    assert actual_hello["Extensions"][0] == {
        "ExtensionType": "extended_master_secret (23)",
    }

    assert actual_hello["Extensions"][1]["ExtensionType"] == "renegotiation_info (65281)"
    assert actual_hello["Extensions"][1]["RawData"] == "00"

    assert actual_handshakes[1]["HandshakeType"] == "certificate (11)"
    assert actual_handshakes[1]["RawData"] is not None
    actual_certificate = actual_handshakes[1]["Data"]
    assert actual_certificate["Certificate"] is not None

    assert actual_handshakes[2]["HandshakeType"] == "server_key_exchange (12)"
    assert actual_handshakes[2]["RawData"] is not None
    actual_server_key_exch = actual_handshakes[2]["Data"]
    assert actual_server_key_exch["CurveType"] == "named_curve (3)"
    assert actual_server_key_exch["Curve"] == "secp384r1 (24)"
    assert actual_server_key_exch["PublicKey"] is not None
    assert actual_server_key_exch["SignatureAlgorithm"] is None
    assert actual_server_key_exch["Signature"] is not None

    assert actual_handshakes[3]["HandshakeType"] == "server_hello_done (14)"
    assert actual_handshakes[3]["RawData"] is not None


def test_tls12_client_key_exchange(capsys):
    entrypoint.main(["--token", base64.b64encode(get_data("tls1.2_client_key_exchange")).decode()])
    actual_out = capsys.readouterr()

    assert actual_out.err == ""
    actual = json.loads(actual_out.out)

    assert isinstance(actual, list)
    assert len(actual) == 1

    assert isinstance(actual[0], dict)
    assert actual[0]["ContentType"] == "handshake (22)"
    assert actual[0]["ProtocolVersion"] == "TLS 1.2 (0x0303) (771)"
    assert actual[0]["RawData"] is not None

    actual_handshakes = actual[0]["Data"]
    assert isinstance(actual_handshakes, list)
    assert len(actual_handshakes) == 1
    assert isinstance(actual_handshakes[0], dict)

    assert actual_handshakes[0]["HandshakeType"] == "client_key_exchange (16)"
    assert actual_handshakes[0]["RawData"] is not None

    actual_ckx = actual_handshakes[0]["Data"]
    assert actual_ckx["PublicKey"] is not None
